import abc
import inspect
from typing import Any, ClassVar, Type

from ._abstract_solver import AbstractSolver, SolverState


class SolverAdapter(AbstractSolver[SolverState], abc.ABC):
    """
    Base class for adapters wrapping solvers from JAXopt and Optimistix.

    Needs to define the class attribute `_solver_cls` and set
    the wrapped solver in the `_solver` attribute.
    """

    _solver_cls: ClassVar[Type]
    _solver: Any

    def __getattr__(self, name: str):
        """Try getting undefined attributes from the underlying solver."""
        # without this guard deepcopy leads to a RecursionError
        try:
            solver = object.__getattribute__(self, "_solver")
        except AttributeError:
            raise AttributeError(name)

        return getattr(solver, name)

    @classmethod
    def get_accepted_arguments(cls) -> set[str]:
        """Set of accepted argument names, extended with the wrapped solver's arguments."""
        own_arguments = set(inspect.getfullargspec(cls.__init__).args)
        solver_arguments = set(inspect.getfullargspec(cls._solver_cls).args)
        solver_init_arguments = set(
            inspect.getfullargspec(cls._solver_cls.__init__).args
        )

        all_arguments = own_arguments | solver_arguments | solver_init_arguments

        # discard arguments that are passed by BaseRegressor
        all_arguments.discard("self")
        all_arguments.discard("unregularized_loss")
        all_arguments.discard("regularizer")
        all_arguments.discard("regularizer_strength")

        return all_arguments

    @classmethod
    def _note_about_accepted_arguments(cls) -> str:
        """Add a potential note about the accepted arguments in the docstring."""
        return ""

    def __init_subclass__(cls, **kw):
        """Generate the docstring including accepted arguments and the wrapped solver's documentation."""
        super().__init_subclass__(**kw)

        # can only do anything if there is a _solver_cls class attribute
        solver_cls = getattr(cls, "_solver_cls", None)
        if solver_cls is None:
            return

        # read the class's docstring or set it to a default
        adapter_doc = inspect.cleandoc(
            inspect.getdoc(cls) or f"Adapter for {solver_cls.__name__}"
        )

        # make a list of accepted arguments
        accepted_doc_header = inspect.cleandoc(
            """
            Accepted arguments:
            -------------------
            """
        )
        accepted_doc = "\n".join(f"- {a}" for a in sorted(cls.get_accepted_arguments()))
        accepted_doc = accepted_doc_header + "\n" + accepted_doc

        # potentially add a note about the accepted arguments
        after_accepted = inspect.cleandoc(cls._note_about_accepted_arguments()).strip()
        if after_accepted:
            accepted_doc = accepted_doc + "\n\n" + after_accepted

        # read the underlying solver class's documentation
        solver_doc_header = inspect.cleandoc(
            f"""
            {solver_cls.__name__}'s documentation:
            """
        )
        solver_doc_header += "\n" + "-" * len(solver_doc_header)
        solver_doc = inspect.cleandoc(
            inspect.getdoc(solver_cls) or "No class documentation found."
        )
        solver_doc = solver_doc_header + "\n" + solver_doc

        # read the underlying solver's __init__'s documentation
        solver_init_doc_header = inspect.cleandoc(
            f"""
            More info from {solver_cls.__name__}.__init__
            """
        )
        solver_init_doc_header += "\n" + "-" * len(solver_init_doc_header)
        solver_init_doc = inspect.cleandoc(
            inspect.getdoc(solver_cls.__init__) or "No __init__ documentation found."
        )
        solver_init_doc = solver_init_doc_header + "\n" + solver_init_doc

        # the whole documentation is the parts after each other separated by blank lines
        class_doc = "\n\n".join(
            (
                adapter_doc,
                accepted_doc,
                solver_doc,
                solver_init_doc,
            )
        )

        cls.__doc__ = inspect.cleandoc(class_doc)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest, parameterized

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from nemos.third_party.jaxopt.jaxopt import tree_util
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


class TreeUtilTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()
    rng = onp.random.RandomState(0)

    self.tree_A = (rng.randn(20, 10) + 1j*rng.randn(20, 10), rng.randn(20))
    self.tree_B = (rng.randn(20, 10), rng.randn(20))

    self.tree_A_dict = (1.0, {"k1": 1.0, "k2": (1.0, 1.0)}, 1.0)
    self.tree_B_dict = (1.0, {"k1": 2.0, "k2": (3.0, 4.0)}, 5.0)

    self.array_A = rng.randn(20) + 1j*rng.randn(20)
    self.array_B = rng.randn(20)
    self.array_C = rng.randn(20) + 1j*rng.randn(20)

  def _assertTreesAllClose(self, tree_A, tree_B):
    self.assertArraysAllClose(tree_A[0], tree_B[0])
    self.assertArraysAllClose(tree_A[1], tree_B[1])

  def test_tree_add(self):
    expected = self.array_A + self.array_B
    got = tree_util.tree_add(self.array_A, self.array_B)
    self._assertTreesAllClose(expected, got)

    expected = (self.tree_A[0] + self.tree_B[0],
                self.tree_A[1] + self.tree_B[1])
    got = tree_util.tree_add(self.tree_A, self.tree_B)
    self._assertTreesAllClose(expected, got)

  def test_tree_sub(self):
    expected = self.array_A - self.array_B
    got = tree_util.tree_sub(self.array_A, self.array_B)
    self._assertTreesAllClose(expected, got)

    expected = (self.tree_A[0] - self.tree_B[0],
                self.tree_A[1] - self.tree_B[1])
    got = tree_util.tree_sub(self.tree_A, self.tree_B)
    self._assertTreesAllClose(expected, got)

  def test_tree_mul(self):
    expected = self.array_A * self.array_B
    got = tree_util.tree_mul(self.array_A, self.array_B)
    self._assertTreesAllClose(expected, got)

    expected = (self.tree_A[0] * self.tree_B[0],
                self.tree_A[1] * self.tree_B[1])
    got = tree_util.tree_mul(self.tree_A, self.tree_B)
    self._assertTreesAllClose(expected, got)

  def test_tree_scalar_mul(self):
    expected = 0.5 * self.array_A
    got = tree_util.tree_scalar_mul(0.5, self.array_A)
    self._assertTreesAllClose(expected, got)

    expected = (0.5 * self.tree_A[0], 0.5 * self.tree_A[1])
    got = tree_util.tree_scalar_mul(0.5, self.tree_A)
    self._assertTreesAllClose(expected, got)

  def test_tree_add_scalar_mul(self):
    expected = (self.tree_A[0] + 0.5 * self.tree_B[0],
                self.tree_A[1] + 0.5 * self.tree_B[1])
    got = tree_util.tree_add_scalar_mul(self.tree_A, 0.5, self.tree_B)
    self._assertTreesAllClose(expected, got)

  def test_tree_vdot(self):
    expected = jnp.vdot(self.array_A, self.array_B)
    got = tree_util.tree_vdot(self.array_A, self.array_B)
    self.assertAllClose(expected, got)

    expected = 15.0
    got = tree_util.tree_vdot(self.tree_A_dict, self.tree_B_dict)
    self.assertAllClose(expected, got)

    expected = (jnp.vdot(self.tree_A[0], self.tree_B[0]) +
                jnp.vdot(self.tree_A[1], self.tree_B[1]))
    got = tree_util.tree_vdot(self.tree_A, self.tree_B)
    self.assertAllClose(expected, got)

  def test_tree_vdot_real(self):
    expected = onp.vdot(self.array_A, self.array_B).real
    got = tree_util.tree_vdot_real(self.array_A, self.array_B)
    self.assertAllClose(expected, got)

    expected = onp.vdot(self.array_A, self.array_C).real
    got = tree_util.tree_vdot_real(self.array_A, self.array_C)
    self.assertAllClose(expected, got)

    expected = sum([onp.vdot(self.tree_A[i], self.tree_B[i]).real for i in range(2)])
    got = tree_util.tree_vdot_real(self.tree_A, self.tree_B)
    self.assertAllClose(expected, got)


  def test_tree_div(self):
    expected = (self.tree_A[0] / self.tree_B[0], self.tree_A[1] / self.tree_B[1])
    got = tree_util.tree_div(self.tree_A, self.tree_B)
    self.assertAllClose(expected, got)

    got = tree_util.tree_div(self.tree_A_dict, self.tree_B_dict)
    expected = (1.0, {'k1': 0.5, 'k2': (0.333333333, 0.25)}, 0.2)
    self.assertAllClose(expected, got, atol=1e-8)

  def test_tree_sum(self):
    expected = jnp.sum(self.array_A)
    got = tree_util.tree_sum(self.array_A)
    self.assertAllClose(expected, got)

    expected = (jnp.sum(self.tree_A[0]) + jnp.sum(self.tree_A[1]))
    got = tree_util.tree_sum(self.tree_A)
    self.assertAllClose(expected, got)

  def test_tree_l2_norm(self):
    expected = jnp.sqrt(jnp.vdot(self.array_A, self.array_A).real)
    got = tree_util.tree_l2_norm(self.array_A)
    self.assertAllClose(expected, got)

    expected = jnp.sqrt(jnp.vdot(self.tree_A[0], self.tree_A[0]).real +
                        jnp.vdot(self.tree_A[1], self.tree_A[1]).real)
    got = tree_util.tree_l2_norm(self.tree_A)
    self.assertAllClose(expected, got)

  def test_tree_inf_norm(self):
    expected = jnp.max(jnp.abs(self.array_A))
    got = tree_util.tree_inf_norm(self.array_A)
    self.assertAllClose(expected, got)

    expected = jnp.maximum(jnp.max(jnp.abs(self.tree_A[0])),
                           jnp.max(jnp.abs(self.tree_A[1])))
    got = tree_util.tree_inf_norm(self.tree_A)
    self.assertAllClose(expected, got)

    tree_a_with_empty_leaves = ({'a': self.tree_A, 'b': jnp.array([])},
                                jnp.array([]))
    expected = tree_util.tree_inf_norm(self.tree_A)
    got = tree_util.tree_inf_norm(tree_a_with_empty_leaves)
    self.assertAllClose(expected, got)

  def test_tree_conj(self):
    expected = onp.conj(self.array_A)
    got = tree_util.tree_conj(self.array_A)
    self.assertAllClose(expected, got)

    expected = (onp.conj(self.tree_A[0]), onp.conj(self.tree_A[1]))
    got = tree_util.tree_conj(self.tree_A)
    self.assertAllClose(expected, got)

  def test_tree_real(self):
    expected = onp.real(self.array_A)
    got = tree_util.tree_real(self.array_A)
    self.assertAllClose(expected, got)

    expected = (onp.real(self.tree_A[0]), onp.real(self.tree_A[1]))
    got = tree_util.tree_real(self.tree_A)
    self.assertAllClose(expected, got)

  def test_tree_imag(self):
    expected = onp.imag(self.array_A)
    got = tree_util.tree_imag(self.array_A)
    self.assertAllClose(expected, got)

    expected = (onp.imag(self.tree_A[0]), onp.imag(self.tree_A[1]))
    got = tree_util.tree_imag(self.tree_A)
    self.assertAllClose(expected, got)


  def test_tree_zeros_like(self):
    tree = tree_util.tree_zeros_like(self.tree_A)
    self.assertAllClose(tree_util.tree_l2_norm(tree), 0.0)

  def test_tree_where(self):
    c = jnp.array([True, False, True, False]), jnp.array([True, False, False, True])
    a = jnp.array([1., 2., 3., 4.]), jnp.array([5., 6., 7., 8.])
    b = jnp.array(42.), jnp.array(13.)
    d = tree_util.tree_where(c, a, b)
    self.assertAllClose(d[0], jnp.array([1., 42., 3., 42.]))
    self.assertAllClose(d[1], jnp.array([5., 13., 13., 8.]))

    c = [jnp.array([True, False, True, False])]
    a = [jnp.array([1., 2., 3., 4.])]
    b = [jnp.array(42.)]
    d = tree_util.tree_where(c, a, b)
    self.assertAllClose(d[0], jnp.array([1., 42., 3., 42.]))

    c = jnp.asarray(False)
    a = [jnp.array([1., 2., 3., 4.]), jnp.array([5., 6., 7., 8.])]
    b = jnp.array(42.)
    d = tree_util.tree_where(c, a, b)
    self.assertAllClose(d[0], jnp.array([42., 42., 42., 42.]))
    self.assertAllClose(d[1], jnp.array([42., 42., 42., 42.]))

  def test_broadcast_pytrees(self):
    # No leaf pytrees, treedefs match.
    trees_in = ([jnp.asarray(1.), jnp.asarray(2.)],
                [jnp.asarray(3.), jnp.asarray(4.)])
    trees_out = tree_util.broadcast_pytrees(*trees_in)
    for l_in, l_out in zip(jtu.tree_leaves(trees_in),
                           jtu.tree_leaves(trees_out)):
      self.assertAllClose(l_in, l_out)

    # One leaf pytree. `a_out` should match the structure of `b_in` and `b_out`
    # by broadcasting `a_in`.
    a_in = jnp.asarray(1.)
    b_in = [jnp.asarray(2.), jnp.asarray(3.)]
    a_out, b_out = tree_util.broadcast_pytrees(a_in, b_in)
    self.assertAllClose(a_out[0], a_in)
    self.assertAllClose(a_out[1], a_in)
    self.assertAllClose(b_out[0], b_in[0])
    self.assertAllClose(b_out[1], b_in[1])

    # Treedefs do not match.
    a_in = [jnp.asarray(1.)]
    b_in = [jnp.asarray(2.), jnp.asarray(3.)]
    with self.assertRaises(ValueError):
      a_out, b_out = tree_util.broadcast_pytrees(a_in, b_in)

  @parameterized.product(high_precision=[True, False], convert_in_jax_dtype=[True, False])
  def test_tree_single_dtype(self, high_precision, convert_in_jax_dtype):
    from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_single_dtype
    current_precision = jax.config.jax_enable_x64
    if high_precision:
      jax.config.update("jax_enable_x64", True)
    else:
      jax.config.update("jax_enable_x64", False)

    dtype_conversions = {
      onp.asarray(True).dtype: jnp.asarray(True).dtype,
      onp.asarray(0).dtype: jnp.asarray(0).dtype,
      onp.asarray(0.0).dtype: jnp.asarray(0.0).dtype,
      onp.asarray(1j * 1.0).dtype: jnp.asarray(1j * 1.0).dtype,
    }

    test_base_types = [
      (True, False),
      (1, 2),
      (3.0, 4.0),
      (1.0 + 1j * 2.0, 2.0 + 1j * 3.0),
      (onp.asarray((1.0, 3.0)), onp.asarray((2.0, 4.0))),
      (jnp.asarray((1.0, 2.0)), jnp.asarray((3.0, 4.0))),
    ]
    for x in test_base_types:
      dtype_tree = tree_single_dtype(
        x, convert_in_jax_dtype=convert_in_jax_dtype
      )
      dtype_leaf = onp.asarray(x[0]).dtype
      if not high_precision and convert_in_jax_dtype:
        if dtype_leaf not in (onp.asarray(True).dtype, jnp.asarray(0.0).dtype):
          self.assertNotEqual(dtype_tree, dtype_leaf)
          dtype_leaf = dtype_conversions[dtype_leaf]
          self.assertEqual(dtype_tree, dtype_leaf)
        else:
          self.assertEqual(dtype_tree, dtype_leaf)
      else:
        self.assertEqual(dtype_tree, dtype_leaf)

    self.assertRaises(ValueError, tree_single_dtype, (1., 1))
    class MyInt():
      i: int = 1
    self.assertEqual(None, tree_single_dtype(MyInt()))

    jax.config.update("jax_enable_x64", current_precision)

  def test_get_real_dtype(self):
    from nemos.third_party.jaxopt.jaxopt._src.tree_util import get_real_dtype
    complex_dtype = jnp.asarray(1j * 1.).dtype
    real_dtype = jnp.asarray(1.).dtype
    self.assertEqual(get_real_dtype(complex_dtype), real_dtype)


if __name__ == '__main__':
  absltest.main()



r"""Collection of proximal operators.

Proximal operators are a mathematical tools used to solve non-differentiable optimization
problems or to simplify complex ones.

A classical use-case for proximal operator is that of minimizing a penalized loss function where the
penalization is non-differentiable (Lasso, group Lasso etc.). In proximal gradient algorithms, proximal
operators are used to find the parameters that balance the minimization of the penalty term with
 the proximity to the gradient descent update of the un-penalized loss.

More formally, proximal operators solve the minimization problem,

$$
\\text{prox}_f(\bm{v}) = \arg\min_{\bm{x}} \left( f(\bm{x}) + \frac{1}{2}\Vert \bm{x} - \bm{v}\Vert_2 ^2 \right)
$$


Where $f$ is usually the non-differentiable penalization term, and $\bm{v}$  is the parameter update of the
un-penalized loss function. The first term controls the penalization magnitude, the second the proximity
with the gradient based update.

References
----------
[1]  Parikh, Neal, and Stephen Boyd. *"Proximal Algorithms, ser. Foundations and Trends (r) in Optimization."* (2013).
"""

from typing import Any, Optional, Tuple

import jax
import jax.numpy as jnp
import jax.tree_util as tree_util


def _norm2_masked(weight_neuron: jnp.ndarray, mask: jnp.ndarray) -> jnp.ndarray:
    """Euclidean norm of the group.

    Calculate the Euclidean norm of the weights for a specified group within a
    neuron's feature vector.

    This function computes the norm of elements that are indicated by the mask array.
    If 'mask' were of boolean type, this operation would be equivalent to performing
    `jnp.linalg.norm(weight_neuron[mask], 2)` followed by dividing the result by the
    square root of the sum of the mask (assuming each group has at least 1 feature).

    Parameters
    ----------
    weight_neuron:
        The feature vector for a neuron. Shape (n_features, ).
    mask:
        The mask vector for group. mask[i] = 1, if the i-th element of weight_neuron
        belongs to the group, 0 otherwise. Shape (n_features, ).

    Returns
    -------
    :
        The norm of the weight vector corresponding to the feature in mask.

    Notes
    -----
        The proximal gradient operator is described in Ming at al.[^1], Proposition 1.

    [^1]:
        Yuan, Ming, and Yi Lin. "Model selection and estimation in regression with grouped variables."
        Journal of the Royal Statistical Society Series B: Statistical Methodology 68.1 (2006): 49-67.
    """
    return jnp.linalg.norm(weight_neuron * mask, 2) / jnp.sqrt(mask.sum())


# vectorize the norm function above
# [(n_neurons, n_features), (n_features)] -> (n_neurons, )
_vmap_norm2_masked_1 = jax.vmap(_norm2_masked, in_axes=(0, None), out_axes=0)
# [(n_neurons, n_features), (n_groups, n_features)] -> (n_neurons, n_groups)
_vmap_norm2_masked_2 = jax.vmap(_vmap_norm2_masked_1, in_axes=(None, 0), out_axes=1)


def prox_group_lasso(
    params: Tuple[jnp.ndarray, jnp.ndarray],
    regularizer_strength: float,
    mask: jnp.ndarray,
    scaling: float = 1.0,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    r"""Proximal gradient operator for group Lasso.

    Parameters
    ----------
    params:
        Weights, shape (n_neurons, n_features) or pytree of same; intercept,
        shape (n_neurons, )
    regularizer_strength:
        The regularization hyperparameter.
    mask:
        ND array of 0,1 as float32, feature mask. size (n_groups, n_features)
        or pytree of same.
    scaling:
        The scaling factor for the group-lasso (it will be set
        depending on the step-size).

    Returns
    -------
    :
        The rescaled weights.

    Notes
    -----
    This function implements the proximal operator for a group-Lasso penalization which
    can be derived in analytical form.
    The proximal operator equation are,

    $$
    \text{prox}(\beta_g) = \text{min}_{\beta} \left[ \lambda  \sum_{g=1}^G \Vert \beta_g \Vert_2 +
     \frac{1}{2} \Vert \hat{\beta} - \beta \Vert_2^2
    \right],
    $$
    where $G$ is the number of groups, and $\beta_g$ is the parameter vector
    associated with the $g$-th group.
    The analytical solution[$^{[1]}$](#references). for the beta is,

    $$
    \text{prox}(\beta_g) = \max \left(1 - \frac{\lambda \sqrt{p_g}}{\Vert \hat{\beta}_g \Vert_2},
     0\right) \cdot \hat{\beta}_g,
    $$
    where $p_g$ is the dimensionality of $\beta_g$ and $\hat{\beta}$ is typically the gradient step
    of the un-regularized optimization objective function. It's easy to see how the group-Lasso
    proximal operator acts as a shrinkage factor for the un-penalize update, and the half-rectification
    non-linearity that effectively sets to zero group of coefficients satisfying,
    $$
    \Vert \hat{\beta}_g \Vert_2 \le \frac{1}{\lambda \sqrt{p_g}}.
    $$

    # References
    ------------
    [1] Yuan, Ming, and Yi Lin. "Model selection and estimation in regression with grouped variables."
    Journal of the Royal Statistical Society Series B: Statistical Methodology 68.1 (2006): 49-67.

    """
    weights, intercepts = params
    shape = weights.shape
    # divide the reg strength by the number of neurons
    regularizer_strength /= intercepts.shape[0]
    # add an extra dim if not 2D, do nothing otherwise.
    weights = jnp.atleast_2d(weights.T)
    # [(n_neurons, n_features), (n_groups, n_features)] -> (n_neurons, n_groups)
    l2_norm = _vmap_norm2_masked_2(weights, mask)
    factor = 1 - regularizer_strength * scaling / l2_norm
    factor = jax.nn.relu(factor)
    # Avoid shrinkage of features that do not belong to any group
    # by setting the shrinkage factor to 1.
    not_regularized = jnp.outer(jnp.ones(factor.shape[0]), 1 - mask.sum(axis=0))
    return (weights * (factor @ mask + not_regularized)).T.reshape(shape), intercepts


def prox_lasso(x: Any, l1reg: Optional[Any] = None, scaling: float = 1.0) -> Any:
    r"""Proximal operator for the l1 norm, i.e., soft-thresholding operator.

    Minimizes the following function:

    $$
      \underset{y}{\text{argmin}} ~ \frac{1}{2} ||x - y||_2^2
      + \text{scaling} \cdot \text{l1reg} \cdot ||y||_1
    $$

    When `l1reg` is a pytree, the weights are applied coordinate-wise.

    Parameters
    ----------
    x :
        Input pytree.
    l1reg :
        Regularization strength, float or pytree with the same structure as `x`. Default is None.
    scaling :
        A scaling factor. Default is 1.0.

    Returns
    -------
    :
        Output pytree with the same structure as `x`.
    """
    if l1reg is None:
        l1reg = 1.0

    if jnp.isscalar(l1reg):
        l1reg = jax.tree_util.tree_map(lambda y: l1reg * jnp.ones_like(y), x)

    def fun(u, v):
        return jnp.sign(u) * jax.nn.relu(jnp.abs(u) - v * scaling)

    return jax.tree_util.tree_map(fun, x, l1reg)


def prox_elastic_net(
    x: Any, hyperparams: Optional[Tuple[Any, Any]] = None, scaling: float = 1.0
) -> Any:
    r"""Proximal operator for the elastic net.

    .. math::

      \underset{y}{\text{argmin}} ~ \frac{1}{2} ||x - y||_2^2
      + \text{scaling} \cdot \text{hyperparams[0]} \cdot g(y)

    where :math:`g(y) = ||y||_1 + \text{hyperparams[1]} \cdot 0.5 \cdot ||y||_2^2`.

    Parameters
    ----------
    x :
        Input pytree.
    hyperparams :
        A tuple, where both ``hyperparams[0]`` and ``hyperparams[1]`` can be either floats or pytrees with the same
        structure as ``x``.
    scaling :
        A scaling factor.

    Returns
    -------
    :
        Output pytree, with the same structure as ``x``.
    """
    if hyperparams is None:
        hyperparams = (0.5, 1.0)

    # hyperparams[0] = regularizer_strength*regularizer_ratio
    lam = (
        tree_util.tree_map(lambda y: hyperparams[0] * jnp.ones_like(y), x)
        if isinstance(hyperparams[0], float)
        else hyperparams[0]
    )
    # hyperparams[1] = (1 - regularizer_ratio)/regularizer_ratio
    gam = (
        tree_util.tree_map(lambda y: hyperparams[1] * jnp.ones_like(y), x)
        if isinstance(hyperparams[1], float)
        else hyperparams[1]
    )

    def prox_l1(u, lambd):
        return jnp.sign(u) * jax.nn.relu(jnp.abs(u) - lambd)

    def fun(u, lambd, gamma):
        return prox_l1(u, scaling * lambd) / (1.0 + scaling * lambd * gamma)

    return tree_util.tree_map(fun, x, lam, gam)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Limited-memory BFGS."""

import warnings

from dataclasses import dataclass
from typing import Any, Callable, NamedTuple, Optional, Union

import jax
import jax.numpy as jnp
from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _init_stepsize
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _setup_linesearch
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_single_dtype
from nemos.third_party.jaxopt.jaxopt._src.tree_util import get_real_dtype
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_conj
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_l2_norm
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_map
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_sub
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_sum
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot_real


def select_ith_tree(tree_history, i):
  """Select tree corresponding to ith history entry."""
  return tree_map(lambda a: a[i, ...], tree_history)

def inv_hessian_product(pytree: Any,
                        s_history: Any,
                        y_history: Any,
                        rho_history: jnp.ndarray,
                        gamma: float = 1.0,
                        start: int = 0):
  """Product between an approximate Hessian inverse and a pytree.

  Histories are pytrees of the same structure as `pytree` except
  that the leaves are arrays of shape `(history_size, ...)`, where
  `...` means the same shape as `pytree`'s leaves.

  The notation follows the reference below.

  Args:
    pytree: pytree to multiply with.
    s_history: pytree whose leaves contain parameter residuals,
      i.e., `s[k] = x[k+1] - x[k]`
    y_history: pytree whose leaves contain gradient residuals,
      i.e., `y[k] = g[k+1] - g[k]`.
    rho_history: array containing `rho[k] = 1. / vdot(s[k], y[k])`.
    gamma: scalar to use for the initial inverse Hessian approximation,
      i.e., `gamma * I`.
    start: starting index in the circular buffer.

  Returns:
    Product between approximate Hessian inverse and the pytree

  Reference:
    Jorge Nocedal and Stephen Wright.
    Numerical Optimization, second edition.
    Algorithm 7.4 (page 178).
  """

  history_size = rho_history.shape[0]

  indices = (start + jnp.arange(history_size)) % history_size

  def body_right(r, i):
    si, yi = select_ith_tree(s_history, i), select_ith_tree(y_history, i)
    alpha = rho_history[i] * tree_vdot_real(si, r)
    r = tree_add_scalar_mul(r, -alpha, yi)
    return r, alpha
  r, alpha = jax.lax.scan(body_right, pytree, indices, reverse=True)

  r = tree_scalar_mul(gamma, r)

  def body_left(r, args):
    i, alpha = args
    si, yi = select_ith_tree(s_history, i), select_ith_tree(y_history, i)
    beta = rho_history[i] * tree_vdot_real(yi, r)
    r = tree_add_scalar_mul(r, alpha - beta, si)
    return r, beta
  r, _ = jax.lax.scan(body_left, r, (indices, alpha))

  return r


def compute_gamma(s_history: Any, y_history: Any, last: int):
  """Compute scalar gamma defining the initialization of the approximate Hessian."""
  # Let gamma = vdot(y_history[last], s_history[last]) / sqnorm(y_history[last]).
  # The initial inverse Hessian approximation can be set to gamma * I.
  # See Numerical Optimization, second edition, equation (7.20).
  # Note that unlike BFGS, the initialization can change on every iteration.

  fun = lambda s_history, y_history: tree_vdot_real(y_history[last], s_history[last])
  num = tree_sum(tree_map(fun, s_history, y_history))

  fun = lambda y_history: tree_vdot_real(y_history[last], y_history[last])
  denom = tree_sum(tree_map(fun, y_history))

  return jnp.where(denom > 0, num / denom, 1.0)


def init_history(pytree, history_size):
  fun = lambda leaf: jnp.zeros((history_size,) + leaf.shape, dtype=leaf.dtype)
  return tree_map(fun, pytree)


def update_history(history_pytree, new_pytree, last):
  fun = lambda history_array, new_value: history_array.at[last].set(new_value)
  return tree_map(fun, history_pytree, new_pytree)


class LbfgsState(NamedTuple):
  """Named tuple containing state information."""
  iter_num: int
  value: float
  grad: Any
  stepsize: float
  error: float
  s_history: Any
  y_history: Any
  rho_history: jnp.ndarray
  gamma: jnp.ndarray
  aux: Optional[Any] = None
  failed_linesearch: bool = False
  num_fun_eval: int = 0
  num_grad_eval: int = 0
  num_linesearch_iter: int = 0


@dataclass(eq=False)
class LBFGS(base.IterativeSolver):
  """LBFGS solver.

  Supports complex variables, see second reference.

  Attributes:
    fun: a smooth function of the form ``fun(x, *args, **kwargs)``.
    value_and_grad: whether ``fun`` just returns the value (False) or both the
      value and gradient (True).
    has_aux: whether ``fun`` outputs auxiliary data or not. If ``has_aux`` is
      False, ``fun`` is expected to be scalar-valued. If ``has_aux`` is True,
      then we have one of the following two cases. If ``value_and_grad`` is
      False, the output should be ``value, aux = fun(...)``. If ``value_and_grad
      == True``, the output should be ``(value, aux), grad = fun(...)``. At each
      iteration of the algorithm, the auxiliary outputs are stored in
      ``state.aux``.
    maxiter: maximum number of proximal gradient descent iterations.
    tol: tolerance of the stopping criterion.
    stepsize: a stepsize to use (if <= 0, use backtracking line search), or a
      callable specifying the **positive** stepsize to use at each iteration.
    linesearch: the type of line search to use: "backtracking" for backtracking
      line search, "zoom" for zoom line search or "hager-zhang" for Hager-Zhang
      line search.
    linesearch_init: strategy for line-search initialization. By default, it
      will use "increase", which will increase the step-size by a factor of
      `increase_factor` at each iteration if the step-size is larger than
      `min_stepsize`, and set it to `max_stepsize` otherwise. Other choices are
      "max", that initializes the step-size to `max_stepsize` at every
      iteration, and "current", that uses the step-size from the previous
      iteration.
    stop_if_linesearch_fails: whether to stop iterations if the line search
      fails. When True, this matches the behavior of core JAX.
    condition: Deprecated. Condition used to select the stepsize when using
      backtracking linesearch.
    maxls: maximum number of iterations to use in the line search.
    decrease_factor: Deprecated. Factor by which to decrease the stepsize during
      line search when using backtracking linesearch (default: 0.8).
    increase_factor: factor by which to increase the stepsize during line search
      (default: 1.5).
    max_stepsize: upper bound on stepsize.
    min_stepsize: lower bound on stepsize guess at start of each linesearch run.
    history_size: size of the memory to use.
    use_gamma: whether to initialize the inverse Hessian approximation with
      gamma * I, where gamma is chosen following equation (7.20) of 'Numerical
      Optimization' (reference below). If use_gamma is set to False, the
      identity is used as initialization.
    implicit_diff: whether to enable implicit diff or autodiff of unrolled
      iterations.
    implicit_diff_solve: the linear system solver to use.
    jit: whether to JIT-compile the optimization loop (default: True).
    unroll: whether to unroll the optimization loop (default: "auto").
    verbose: if set to True or 1 prints the information at each step of 
      the solver, if set to 2, print also the information of the linesearch.

  References:
    Jorge Nocedal and Stephen Wright.
    Numerical Optimization, second edition.
    Algorithm 7.5 (page 179).

    Laurent Sorber, Marc van Barel, and Lieven de Lathauwer.
    Unconstrained Optimization of Real Functions in Complex Variables.
    SIAM J. Optim., Vol. 22, No. 3, pp. 879-898
  """

  fun: Callable
  value_and_grad: bool = False
  has_aux: bool = False

  maxiter: int = 500
  # FIXME: should depend on whether float32 or float64 is used.
  # Tests should pass in float64 without modifying tol
  tol: float = 1e-3

  stepsize: Union[float, Callable] = 0.0
  linesearch: str = "zoom"
  linesearch_init: str = "increase"
  stop_if_linesearch_fails: bool = False
  condition: Any = None  # deprecated in v0.8
  maxls: int = 30
  decrease_factor: Any = None  # deprecated in v0.8
  increase_factor: float = 1.5
  max_stepsize: float = 1.0
  # FIXME: should depend on whether float32 or float64 is used.
  min_stepsize: float = 1e-6

  history_size: int = 10
  use_gamma: bool = True

  implicit_diff: bool = True
  implicit_diff_solve: Optional[Callable] = None

  jit: bool = True
  unroll: base.AutoOrBoolean = "auto"

  verbose: Union[bool, int] = False

  def _cond_fun(self, inputs):
    _, state = inputs[0]
    # We continue the optimization loop while the error tolerance is not met and,
    # either failed linesearch is disallowed or linesearch hasn't failed.
    return (state.error > self.tol) & jnp.logical_or(not self.stop_if_linesearch_fails, ~state.failed_linesearch)

  def init_state(self,
                 init_params: Any,
                 *args,
                 **kwargs) -> LbfgsState:
    """Initialize the solver state.

    Args:
      init_params: pytree containing the initial parameters.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      state
    """
    if isinstance(init_params, base.OptStep):
      # `init_params` can either be a pytree or an OptStep object
      state_kwargs = dict(
        s_history=init_params.state.s_history,
        y_history=init_params.state.y_history,
        rho_history=init_params.state.rho_history,
        gamma=init_params.state.gamma,
        iter_num=init_params.state.iter_num,
        stepsize=init_params.state.stepsize,
      )
      init_params = init_params.params
      realdtype = get_real_dtype(tree_single_dtype(init_params))
    else:
      realdtype = get_real_dtype(tree_single_dtype(init_params))
      state_kwargs = dict(
          s_history=init_history(init_params, self.history_size),
          y_history=init_history(init_params, self.history_size),
          rho_history=jnp.zeros(self.history_size, dtype=realdtype),
          gamma=jnp.asarray(1.0, dtype=realdtype),
          iter_num=jnp.asarray(0),
          stepsize=jnp.asarray(self.max_stepsize, dtype=realdtype),
      )
    (value, aux), grad = self._value_and_grad_with_aux(init_params, *args, **kwargs)
    return LbfgsState(value=value,
                      grad=grad,
                      error=jnp.asarray(jnp.inf, dtype=realdtype),
                      **state_kwargs,
                      aux=aux,
                      failed_linesearch=jnp.asarray(False),
                      num_grad_eval=jnp.array(1, base.NUM_EVAL_DTYPE),
                      num_fun_eval=jnp.array(1, base.NUM_EVAL_DTYPE),
                      num_linesearch_iter=jnp.array(0, base.NUM_EVAL_DTYPE))

  def update(self,
             params: Any,
             state: LbfgsState,
             *args,
             **kwargs) -> base.OptStep:
    """Performs one iteration of LBFGS.

    Args:
      params: pytree containing the parameters.
      state: named tuple containing the solver state.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      (params, state)
    """
    if isinstance(params, base.OptStep):
      params = params.params

    realdtype = state.rho_history.dtype  # avoid recomputation, take realdtype from initialized state

    start = state.iter_num % self.history_size
    value, grad = (state.value, state.grad)
    descent_direction = tree_scalar_mul(-1.0, tree_conj(grad))
    s_history = state.s_history
    y_history = state.y_history
    rho_history = state.rho_history

    if self.history_size:
      descent_direction = inv_hessian_product(
          pytree=descent_direction,
          s_history=s_history,
          y_history=y_history,
          rho_history=rho_history,
          gamma=state.gamma,
          start=start,
      )

    use_linesearch = (
        not isinstance(self.stepsize, Callable) and self.stepsize <= 0
    )
    if use_linesearch:
      init_stepsize = _init_stepsize(
          self.linesearch_init,
          self.max_stepsize,
          self.min_stepsize,
          self.increase_factor,
          state.stepsize,
      )
      new_stepsize, ls_state = self.run_ls(
          init_stepsize,
          params,
          value=value,
          grad=grad,
          descent_direction=descent_direction,
          fun_args=args,
          fun_kwargs=kwargs)
      new_params = ls_state.params
      new_value = ls_state.value
      new_grad = ls_state.grad
      new_aux = ls_state.aux
      failed_linesearch = ls_state.failed
      new_num_linesearch_iter = state.num_linesearch_iter + ls_state.iter_num
      new_num_grad_eval = state.num_grad_eval + ls_state.num_grad_eval
      new_num_fun_eval = state.num_fun_eval + ls_state.num_fun_eval

    else:
      if isinstance(self.stepsize, Callable):
        new_stepsize = self.stepsize(state.iter_num)
      else:
        new_stepsize = self.stepsize

      new_params = tree_add_scalar_mul(params, new_stepsize, descent_direction)
      (new_value, new_aux), new_grad = self._value_and_grad_with_aux(
          new_params, *args, **kwargs
      )
      new_num_grad_eval = state.num_grad_eval + 1
      new_num_fun_eval = state.num_fun_eval + 1
      new_num_linesearch_iter = state.num_linesearch_iter
      failed_linesearch = jnp.asarray(False)
    s = tree_sub(new_params, params)
    y = tree_conj(tree_sub(new_grad, grad))
    vdot_sy = tree_vdot_real(s, y)
    rho = jnp.where(vdot_sy == 0, 0, 1. / vdot_sy)

    if self.history_size:
      s_history = update_history(s_history, s, start)
      y_history = update_history(y_history, y, start)
      rho_history = update_history(rho_history, rho, start)

    if self.history_size and self.use_gamma:
      gamma = compute_gamma(s_history, y_history, start)
    else:
      gamma = jnp.array(1.0, dtype=realdtype)

    error = tree_l2_norm(new_grad)
    new_state = LbfgsState(iter_num=state.iter_num + 1,
                           value=new_value,
                           grad=new_grad,
                           stepsize=jnp.asarray(new_stepsize, dtype=realdtype),
                           error=jnp.asarray(error, dtype=realdtype),
                           s_history=s_history,
                           y_history=y_history,
                           rho_history=rho_history,
                           gamma=gamma,
                           aux=new_aux,
                           failed_linesearch=failed_linesearch,
                           num_grad_eval=new_num_grad_eval,
                           num_fun_eval=new_num_fun_eval,
                           num_linesearch_iter=new_num_linesearch_iter)

    if self.verbose:
      self.log_info(
          new_state,
          error_name="Gradient Norm",
          additional_info={
              "Objective Value": new_value,
              "Stepsize": new_stepsize,
              "Number Linesearch Iterations": 
              new_state.num_linesearch_iter - state.num_linesearch_iter
          }
      )
    return base.OptStep(params=new_params, state=new_state)

  def optimality_fun(self, params, *args, **kwargs):
    """Optimality function mapping compatible with ``@custom_root``."""
    return self._value_and_grad_fun(params, *args, **kwargs)[1]

  def _value_and_grad_fun(self,
                          params,
                          *args,
                          **kwargs):
    if isinstance(params, base.OptStep):
      params = params.params
    (value, _), grad = self._value_and_grad_with_aux(params, *args, **kwargs)
    return value, grad

  def __post_init__(self):
    super().__post_init__()

    _fun_with_aux, _, self._value_and_grad_with_aux = \
      base._make_funs_with_aux(fun=self.fun,
                               value_and_grad=self.value_and_grad,
                               has_aux=self.has_aux)

    self.reference_signature = self.fun

    unroll = self._get_unroll_option()

    self.linesearch_solver = _setup_linesearch(
        linesearch=self.linesearch,
        fun=_fun_with_aux,
        value_and_grad=self._value_and_grad_with_aux,
        has_aux=True,
        maxlsiter=self.maxls,
        max_stepsize=self.max_stepsize,
        jit=self.jit,
        unroll=unroll,
        verbose=max(int(self.verbose)-1, 0),
    )
    self.run_ls = self.linesearch_solver.run

    # FIXME: to remove in future releases
    if self.condition is not None:
      warnings.warn("Argument condition is deprecated", DeprecationWarning)
    if self.decrease_factor is not None:
      warnings.warn(
          "Argument decrease_factor is deprecated", DeprecationWarning
      )



"""Utility functions for data pre-processing."""

import inspect
import os
import warnings
from importlib.metadata import version
from typing import Any, Callable, List, Literal, Optional, Union

import jax
import jax.numpy as jnp
import numpy as np
from numpy.typing import NDArray

from .base_class import Base
from .tree_utils import pytree_map_and_reduce
from .type_casting import is_at_least_1d_numpy_array_like, support_pynapple
from .typing import Pytree

__all__ = [
    "check_dimensionality",
    "validate_axis",
    "nan_pad",
    "shift_time_series",
    "row_wise_kron",
    "one_over_x",
    "get_flattener_unflattener",
]


def __dir__() -> list[str]:
    return __all__


SPECIAL_KEY_NAMES = {
    jax.scipy.stats.norm.cdf: "norm.cdf",
}


def check_dimensionality(
    pytree: Any,
    expected_dim: int,
) -> bool:
    """
    Check the dimensionality of the arrays in a pytree.

    Check that all arrays in pytree have the expected dimensionality.

    Parameters
    ----------
    pytree :
        A pytree object.
    expected_dim :
        Number of expected dimension for the NDArrays.

    Returns
    -------
    True if all the arrays has the expected number of dimension, False otherwise.
    """
    return not pytree_map_and_reduce(lambda x: x.ndim != expected_dim, any, pytree)


def validate_axis(pytree: Any, axis: int):
    """
    Validate the axis for each array in a given tree structure.

    This function checks if the specified axis exists in each array within the tree. It raises a ValueError
    if the specified axis is equal to or greater than the number of dimensions in any of the arrays.

    Parameters
    ----------
    pytree :
        A tree containing arrays.
    axis :
        The axis that should be valid for each array in the tree. This means each array must have at least
        `axis + 1` dimensions.

    Raises
    ------
    ValueError
        If the specified axis is equal to or greater than the number of dimensions (`ndim`) of any array
        within the tree. This ensures that operations intended for a specific axis can be safely performed
        on every array in the tree.
    ValueError
        If the axis is negative or non-integer.
    """
    if not isinstance(axis, int) or axis < 0:
        raise ValueError("`axis` must be a non negative integer.")

    if pytree_map_and_reduce(lambda x: x.ndim <= axis, any, pytree):
        raise ValueError(
            "'axis' must be smaller than the number of dimensions of any array in 'tree'."
        )


def check_non_empty(pytree: Any, pytree_name: str):
    """
    Check if any array in the pytree is empty.

    Parameters
    ----------
    pytree :
        A pytree object containing arrays.
    pytree_name :
        The name of the pytree variable for error message purposes.

    Raises
    ------
    ValueError
        If any array in the pytree is empty (i.e., has a zero dimension).
    """
    if pytree_map_and_reduce(lambda x: 0 in x.shape, any, pytree):
        raise ValueError(
            f"Empty array provided. At least one of dimension in {pytree_name} is empty."
        )


def _pad_dimension(
    array: jnp.ndarray,
    axis: int,
    pad_size: int,
    predictor_causality: Literal["causal", "acausal", "anti-causal"] = "causal",
    constant_values: float = jnp.nan,
) -> jnp.ndarray:
    """
    Add padding to the last dimension of an array based on the convolution type.

    This is a helper function used by `nan_pad_conv`, which is the function we expect the user will interact with.

    Parameters
    ----------
    array:
        The array to be padded.
    axis:
        The axis to be padded.
    pad_size:
        The number of NaNs to concatenate as padding.
    predictor_causality:
        Causality of this predictor, which determines where padded values are added.
    constant_values:
        The constant values for padding, default is jnp.nan.

    Returns
    -------
    :
        An array with padded last dimension.
    """
    padding_settings = {
        "causal": (pad_size, 0),
        "acausal": ((pad_size) // 2, pad_size - (pad_size) // 2),
        "anti-causal": (0, pad_size),
    }
    # convert negative axis in jax jit compilable way
    axis = axis * (axis >= 0) + (array.ndim + axis) * (axis < 0)

    pad_width = (
        ((0, 0),) * axis
        + (padding_settings[predictor_causality],)
        + ((0, 0),) * (array.ndim - 1 - axis)
    )
    return jnp.pad(array, pad_width, constant_values=constant_values)


def nan_pad(
    conv_time_series: Any,
    pad_size: int,
    predictor_causality: Literal["causal", "acausal", "anti-causal"] = "causal",
    axis: int = 0,
) -> Any:
    """
    Add NaN padding to a convolved time series based on specified causality and axis.

    This function pads the convolved time series with NaNs along a specified axis. The amount
    and location of the padding are determined by the pad_size and predictor_causality parameters.

    Parameters
    ----------
    conv_time_series :
        The convolved time series to pad. This variable should be a pytree with arrays as leaves.
        The structure can be one of the following:
        1. A single array with ndim > axis.
        2. A pytree whose leaves are arrays.
    pad_size :
        The number of NaNs to concatenate as padding.
    predictor_causality : {'causal', 'acausal', 'anti-causal'}, default='causal'
        Causality of the predictor, which determines where padded values are added:
        - 'causal': Padding is added before the data.
        - 'acausal': Padding is evenly distributed before and after the data.
        - 'anti-causal': Padding is added after the data.
    axis : int, default=0
        The axis along which to add padding.

    Returns
    -------
    padded_conv_time_series : Any
        The convolved time series with NaN padding. The structure matches that of `conv_time_series`.

    Raises
    ------
    ValueError
        - If `pad_size` is not a positive integer.
        - If `predictor_causality` is not one of the expected values ('causal', 'acausal', 'anti-causal').
        - If `axis` is not a valid axis for any of the arrays in `conv_time_series`, specifically
          if `axis >= array.ndim` for any array.
        - If any array in `conv_time_series` does not have a floating-point data type.
    """
    if not isinstance(pad_size, int) or pad_size <= 0:
        raise ValueError(
            f"pad_size must be a positive integer! Pad size of {pad_size} provided instead!"
        )

    causality_choices = ["causal", "acausal", "anti-causal"]
    if predictor_causality not in causality_choices:
        raise ValueError(
            f"predictor_causality must be one of {causality_choices}. {predictor_causality} provided instead!"
        )
    if predictor_causality == "acausal" and (pad_size % 2 == 1):
        warnings.warn(
            "With acausal filter, pad_size should probably be even,"
            " so that we can place an equal number of NaNs on either side of input",
            UserWarning,
        )

    # convert to jax ndarray
    conv_time_series = jax.tree_util.tree_map(jnp.asarray, conv_time_series)

    # validate the axis
    validate_axis(conv_time_series, axis)

    if is_at_least_1d_numpy_array_like(conv_time_series):
        if not np.issubdtype(conv_time_series.dtype, np.floating):
            raise ValueError("conv_time_series must have a float dtype!")
        return _pad_dimension(
            conv_time_series,
            axis,
            pad_size,
            predictor_causality,
            constant_values=jnp.nan,
        )

    else:
        if pytree_map_and_reduce(
            lambda trial: not np.issubdtype(trial.dtype, np.floating),
            any,
            conv_time_series,
        ):
            raise ValueError("All leaves of conv_time_series must have a float dtype!")
        return jax.tree_util.tree_map(
            lambda trial: _pad_dimension(
                trial, axis, pad_size, predictor_causality, constant_values=jnp.nan
            ),
            conv_time_series,
        )


def _compute_index_adjust(
    time_series: NDArray, causality: Literal["causal", "anti-causal"], axis: int
):
    """Compute index adjustment for shifting a time series."""
    adjust_indices = {
        "causal": (0, time_series.shape[axis] - 1),
        "anti-causal": (1, time_series.shape[axis]),
    }
    return adjust_indices[causality]


def shift_time_series(
    time_series: Any,
    predictor_causality: Literal["causal", "anti-causal"] = "causal",
    axis: int = 0,
):
    """Shift time series based on causality of predictor, adding NaNs as needed.

    Shift a time series based on the causality of the predictor and adds NaNs as needed,
    with the operation applied along a specified axis.

    Parameters
    ----------
    time_series :
        The time series to shift, which can be a single array or a pytree of arrays.
        Each array should have a floating-point data type.
    predictor_causality :
        Determines the direction of the shift:
        - 'causal': Shifts the series forward, inserting a NaN at the start.
        - 'anti-causal': Shifts the series backward, appending a NaN at the end.
    axis :
        The axis along which to perform the shift. Must be valid for all arrays in the time series.

    Returns
    -------
    shifted_time_series : Any
        The shifted time series. The structure matches that of `time_series`, with each element
        shifted according to `predictor_causality` and NaNs added accordingly.

    Raises
    ------
    ValueError
        - If `predictor_causality` is not 'causal' or 'anti-causal'.
        - If `axis` is invalid for any array within `time_series`.
        - If any array in `time_series` does not have a floating-point data type.

    Notes
    -----
    The direction of the shift depends on the value of `predictor_causality`:

    - `'causal'`: shift `time_series` one time bin forward and drop final time
      point, so that e.g., `[0, 1, 2]` becomes `[np.nan, 0, 1]`
    - `'anti-causal'`: shift `time_series` one time bin backwards and drop
      first time point, so that e.g., `[0, 1, 2]` becomes `[1, 2, np.nan]`

    """
    # validate axis
    validate_axis(time_series, axis)

    if predictor_causality not in ["causal", "anti-causal"]:
        raise ValueError(
            f"predictor_causality must be one of 'causal', 'anti-causal'. {predictor_causality} provided instead!"
        )

    # compute the start, end indices tree
    adjust_idx = jax.tree_util.tree_map(
        lambda x: _compute_index_adjust(x, predictor_causality, axis), time_series
    )

    # convert to jax ndarray
    time_series = jax.tree_util.tree_map(jnp.asarray, time_series)

    if is_at_least_1d_numpy_array_like(time_series):
        if not np.issubdtype(time_series.dtype, np.floating):
            raise ValueError("time_series must have a float dtype!")
        return _pad_dimension(
            jnp.take(time_series, jnp.arange(*adjust_idx), axis=axis),
            axis,
            1,
            predictor_causality,
            jnp.nan,
        )
    else:
        if pytree_map_and_reduce(
            lambda trial: not np.issubdtype(trial.dtype, np.floating), any, time_series
        ):
            raise ValueError("All leaves of time_series must have a float dtype!")
        return jax.tree_util.tree_map(
            lambda trial, idx: _pad_dimension(
                jnp.take(trial, jnp.arange(*idx), axis=axis),
                axis,
                1,
                predictor_causality,
                jnp.nan,
            ),
            time_series,
            adjust_idx,
        )


def row_wise_kron(
    A: jnp.ndarray, C: jnp.ndarray, jit: bool = False, transpose: bool = True
) -> jnp.ndarray:
    r"""Compute the row-wise Kronecker product.

    Compute the row-wise Kronecker product between two matrices using JAX.
    See [\[1\]](#references) for more details on the Kronecker product.

    Parameters
    ----------
    A : jax.numpy.ndarray
        The first matrix.
    C : jax.numpy.ndarray
        The second matrix.
    jit :
        Activate Just-in-Time (JIT) compilation. Default is False.
    transpose :
        Transpose matrices A and C before computation. Default is True.

    Returns
    -------
    K : jnp.nparray
        The resulting matrix with row-wise Kronecker product.

    Notes
    -----
    This function computes the row-wise Kronecker product between dense matrices A and C
    using JAX for automatic differentiation and GPU acceleration.

    # References
    ------------
    [1] Petersen, Kaare Brandt, and Michael Syskind Pedersen. "The matrix cookbook."
    Technical University of Denmark 7.15 (2008): 510.
    """
    if transpose:
        A = A.T
        C = C.T

    @jax.jit if jit else lambda x: x
    def row_wise_kron(a, c):
        return jnp.kron(a, c)

    K = jax.vmap(row_wise_kron)(A, C)

    if transpose:
        K = K.T

    return K


def assert_has_attribute(obj: Any, attr_name: str):
    """Ensure the object has the given attribute."""
    if not hasattr(obj, attr_name):
        raise AttributeError(
            f"The provided object does not have the required `{attr_name}` attribute!"
        )


def assert_is_callable(func: Callable, func_name: str):
    """Ensure the provided function is callable."""
    if not callable(func):
        raise TypeError(f"The `{func_name}` must be a Callable!")


def assert_returns_ndarray(
    func: Callable, inputs: Union[List[jnp.ndarray], List[float]], func_name: str
):
    """Ensure the function returns a jax.numpy.ndarray."""
    array_out = func(*inputs)
    if not isinstance(array_out, jnp.ndarray):
        raise TypeError(f"The `{func_name}` must return a jax.numpy.ndarray!")


def assert_differentiable(func: Callable, func_name: str):
    """Ensure the function is differentiable."""
    try:
        gradient_fn = jax.grad(func)
        gradient_fn(jnp.array(1.0))
    except Exception as e:
        raise TypeError(f"The `{func_name}` is not differentiable. Error: {str(e)}")


def assert_preserve_shape(
    func: Callable, inputs: List[jnp.ndarray], func_name: str, input_index: int
):
    """Check that the function preserve the input shape."""
    result = func(*inputs)
    if not result.shape == inputs[input_index].shape:
        raise ValueError(f"The `{func_name}` must preserve the input array shape!")


def assert_scalar_func(func: Callable, inputs: List[jnp.ndarray], func_name: str):
    """Check that `func` return an array containing a single scalar."""
    assert_returns_ndarray(func, inputs, func_name)
    array_out = func(*inputs)
    try:
        float(array_out)
    except TypeError:
        raise TypeError(
            f"The `{func_name}` should return a scalar! "
            f"Array of shape {array_out.shape} returned instead!"
        )


def format_repr(
    obj: Base,
    exclude_keys: Optional[List[str]] = None,
    use_name_keys: Optional[List[str]] = None,
    multiline=False,
):
    """
    Format the representation string of an object (`__repr__`).

    This function generates a string representation of an object, including
    the parameters returned by `obj.get_params(deep=False)`. The output excludes
    specified keys and formats certain keys using their class names instead of
    their default string representations.

    Parameters
    ----------
    obj :
        The object inheriting from  Base whose representation string is being formatted.
    exclude_keys :
        List of parameter keys to exclude from the representation. If `None`, no
        keys are excluded. Defaults to `None`.
    use_name_keys :
        List of keys for which the value's `__name__` attribute is used instead
        of the default `__repr__` output. Defaults to an empty list.
    multiline:
        If True, add each parameter on a new line, if False (default), list all parameters
        in a single line.

    Returns
    -------
    :
        A formatted representation string for the object, including its class
        name and relevant parameters.

    Notes
    -----
    - The function includes only non-empty or falsey values (except 0 and False) in the
      representation. The falsey values include empty list or dict, or similar.
    - Parameters are displayed in the order defined in the `obj.__init__`
      method.
    - Parameters with a `shape` attribute (arrays) are excluded from the representation.

    Examples
    --------
    >>> import jax
    >>> from nemos.base_class import Base
    >>> from nemos.regularizer import Ridge
    >>> from nemos.utils import format_repr
    >>> class Example(Base):
    ...     def __init__(self, a, b, c, d):
    ...         self.a = a
    ...         self.b = b
    ...         self.c = c
    ...         self.d = d
    >>> obj = Example(1, jax.numpy.exp, c="hi", d=None)
    >>> format_repr(obj, exclude_keys=["c"], use_name_keys=["b"])
    'Example(a=1, b=exp)'
    """
    exclude_keys = [] if exclude_keys is None else exclude_keys
    use_name_keys = [] if use_name_keys is None else use_name_keys

    init_params = list(inspect.signature(obj.__init__).parameters.keys())
    disp_params = []
    # use special method for basis
    all_params = obj.get_params(deep=False)
    label = all_params.pop("label", None)
    for k, v in all_params.items():
        repr_param = (
            k not in exclude_keys and not hasattr(v, "shape") and (v or v in (0, False))
        )
        if repr_param:
            if k in use_name_keys:
                if v in SPECIAL_KEY_NAMES:
                    v = SPECIAL_KEY_NAMES[v]
                else:
                    v = getattr(v, "__name__", repr(v))
            elif isinstance(v, str):
                v = repr(v)
            disp_params.append(f"{k}={v}")
    disp_params = sorted(disp_params, key=lambda x: init_params.index(x.split("=")[0]))
    cls_name = obj.__class__.__name__
    # if label doesn't exist or is the same as the class name (as is the default for
    # basis), then don't use it
    disp_label = (label is not None) and (label != cls_name)
    if multiline:
        tab = "    "
        disp_params = "\n" + tab + f",\n{tab}".join(disp_params) + "\n" + tab[:-4]
        repr_str = (
            f"{repr(label)}: {cls_name}({disp_params})"
            if disp_label
            else f"{cls_name}({disp_params})"
        )
    else:
        disp_params = ", ".join(disp_params)
        repr_str = (
            f"{repr(label)}: {cls_name}({disp_params})"
            if disp_label
            else f"{cls_name}({disp_params})"
        )
    return repr_str


# enable concatenation for pynapple objects.
pynapple_concatenate_jax = support_pynapple(conv_type="jax")(jnp.concatenate)
pynapple_concatenate_numpy = support_pynapple(conv_type="numpy")(np.concatenate)


def _get_terminal_size():
    """Get the terminal size for __repr__.

    Returns
    -------
    rows:
        Number of rows in the terminal window.
    cols:
        Number of columns in the terminal window.

    """
    cols = 100  # Default
    rows = 2
    try:
        cols, rows = os.get_terminal_size()
    except Exception:
        import shutil

        cols, rows = shutil.get_terminal_size()

    return cols, rows


def one_over_x(x: NDArray):
    """Implement 1/x."""
    return jnp.power(x, -1)


def _flatten_dict(nested_dict: dict, parent_key: str = "") -> dict:
    """
    Flatten a nested dictionary into a single-level dictionary with keys representing the hierarchy.

    Parameters
    ----------
    nested_dict :
        The dictionary to flatten.
    parent_key :
        This key starts blank, but recursively it will be filled with the parent key,
        which is used to create the hierarchy in the flattened dictionary.

    Returns
    -------
    dict :
        A flattened dictionary where the hierarchy is represented by concatenated keys (using __ as a separator).
    """

    sep = "__"
    items = []
    # Iterate over key-value pairs in the dictionary
    for k, v in nested_dict.items():
        # containers (tuple, dict, list for our purposes are labeled)
        if isinstance(v, dict):
            container_label = "dict"
        elif isinstance(v, list):
            container_label = "list"
        elif isinstance(v, tuple):
            container_label = "tuple"
        else:
            container_label = "item"

        new_key = (
            f"{parent_key}{sep}{container_label}{sep}{k}"
            if parent_key
            else f"{container_label}{sep}{k}"
        )

        # Recursively flatten if the value is a dictionary
        if isinstance(v, dict):
            items.extend(_flatten_dict(v, new_key).items())
        # if tuple or list, convert to dict using item index and recursively flatten
        elif isinstance(v, (list, tuple)):
            items.extend(
                _flatten_dict({f"{i}": vi for i, vi in enumerate(v)}, new_key).items()
            )
        else:
            # None values and non-standard types are converted to numpy
            if v is None:
                v = np.nan
            elif not isinstance(v, (str, int, float, bool)):
                v = np.array(v)
            items.append((new_key, v))
    return dict(items)


def _unflatten_dict(flat_dict: dict) -> dict:
    """
    Unflatten a dictionary that was flattened using _flatten_dict, and saved as npz.

    Note that all leaves are expected to be numpy arrays.

    Parameters
    ----------
    flat_dict :
        The flattened dictionary with keys containing type information and hierarchy and leaves numpy arrays.

    Returns
    -------
    dict :
        The reconstructed nested dictionary with original structure and types.
    """
    sep = "__"
    root = {}

    for key, value in flat_dict.items():
        # Replace np.nan back to None
        # Convert numpy string, int, float or nan to their respective types
        if value.ndim == 0:
            if value.dtype.type is np.str_:
                value = str(value)
            elif value.dtype.type is np.int_:
                value = int(value)
            elif issubclass(value.dtype.type, np.floating):
                value = None if np.isnan(value) else float(value)

        # split correctly `varname_` type of params
        parts = [part[::-1] for part in key[::-1].split(sep)[::-1]]

        current = root

        # Walk the path - but handle the last pair differently if it's an 'item'
        pairs = list(zip(parts[::2], parts[1::2]))

        # Navigate through all pairs except potentially the last one
        for i, (container_type, container_key) in enumerate(pairs):
            is_last = i == len(pairs) - 1

            # If this is the last pair and it's an 'item', just set the value directly
            if is_last and container_type == "item":
                current[container_key] = value
                break

            # Otherwise, create/navigate into the container
            if container_key not in current:
                current[container_key] = {"__container_type__": container_type}

            current = current[container_key]
        else:
            # This else clause runs if we didn't break (i.e., no 'item' leaf)
            # This shouldn't happen with your flatten structure, but just in case
            final_key = parts[-1]
            current[final_key] = value

    return reconstruct_object_from_structured_dict(root)


# Convert to proper types
def reconstruct_object_from_structured_dict(obj):
    if not isinstance(obj, dict):
        return obj

    container_type = obj.pop("__container_type__", None)

    # Check if numeric keys (list/tuple)
    keys = [k for k in obj.keys() if k != "__container_type__"]
    if keys and all(k.isdigit() for k in keys):
        sorted_items = sorted(obj.items(), key=lambda x: int(x[0]))
        items = [reconstruct_object_from_structured_dict(v) for k, v in sorted_items]
        return tuple(items) if container_type == "tuple" else items

    # Regular dict
    return {k: reconstruct_object_from_structured_dict(v) for k, v in obj.items()}


def _get_name(x: object) -> str:
    """
    Get the name of an object ``x``, for saving/loading purposes.

    Parameters
    ----------
    x :
        A python object or function.

    Returns
    -------
    name :
        The name of the object, with full module path (e.g.,
        ``nemos.observation_models.PoissonObservations``).
    """
    if x is None:
        return None
    if hasattr(x, "__module__") and hasattr(x, "__name__"):
        # x is a function or class
        return f"{x.__module__}.{x.__name__}"
    elif hasattr(x, "__class__"):
        # x is an instance of a class
        cls = x.__class__
        return f"{cls.__module__}.{cls.__name__}"
    else:
        raise TypeError(f"Cannot retrieve name of variable {x} of type {type(x)}.")


def _is_callable_or_class(obj):
    """Check if obj is callable or class."""
    return callable(obj) or inspect.isclass(obj)


def _unpack_params(params_dict: dict, string_attrs: list = None) -> dict:
    """
    Convert a parameter dictionary into serializable format.

    For objects with `get_params`/`set_params`, extracts the class name and
    parameters. Some attributes are converted to strings to facilitate saving and loading.

    Parameters
    ----------
    params_dict :
        Dictionary of parameters, possibly containing objects.
    string_attrs :
        List of attributes that should be converted to strings (e.g., `inverse_link_function`).

    Returns
    -------
    dict :
        Serializable dictionary with class names and nested parameters.
    """

    out = dict()
    for key, value in params_dict.items():
        # if the parameter is an objet with get_params/set_params,
        # extract its class name and parameters
        if hasattr(value, "get_params") and hasattr(value, "set_params"):
            cls_name = _get_name(value)
            params = _unpack_params(value.get_params(deep=False), string_attrs)
            out[key] = {"class": cls_name, "params": params}
        else:
            # if the parameter is in string_attrs, store its name
            if string_attrs is not None and (
                key in string_attrs or _is_callable_or_class(value)
            ):
                out[key] = _get_name(value)
            else:
                out[key] = value
    return out


def get_env_metadata() -> dict[str, str]:
    """Get environment metadata.

    Get the environment metadata relevant to model fitting.


    Notes
    -----
    ``jax`` and ``jaxlib`` for arrays and linear algebra, ``scipy`` is used at model
    initialization to find numerical inverse for custom link functions, ``scikit-learn``
    is used for pipelines and cross-validation.
    """
    return {
        "jax": version("jax"),
        "jaxlib": version("jaxlib"),
        "scipy": version("scipy"),
        "scikit-learn": version("scikit-learn"),
        "nemos": version("nemos"),
    }


def get_flattener_unflattener(parameter_tree: Pytree):
    """
    Create functions for flattening parameter pytrees and reshaping them to their original shape.

    Parameters
    ----------
    parameter_tree :
        Pytree to flatten. Usually model parameters.

    Returns
    -------
    (flattener, unflattener):
        Tuple of two functions: first one flattens the parameters, second one resshapes the flat ones.
    """
    flat, struct = jax.tree_util.tree_flatten(parameter_tree)
    shapes = [x.shape for x in flat]
    sizes = jnp.array([x.size for x in flat], dtype=int)
    split_indices = jnp.cumsum(sizes[:-1])

    def flattener(parameter_tree):
        flat = jax.tree_util.tree_leaves(parameter_tree)
        return jnp.concatenate([x.flatten() for x in flat])

    def unflattener(flat_params):
        split_params = jnp.split(flat_params, split_indices)
        split_params = [x.reshape(s) for x, s in zip(split_params, shapes)]
        return jax.tree_util.tree_unflatten(struct, split_params)

    return flattener, unflattener



import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
from scipy.special import laguerre

import nemos as nmo
from nemos._inspect_utils.inspect_utils import trim_kwargs

plt.rcParams.update(
    {
        "figure.dpi": 300,
    }
)

KWARGS = dict(
    n_basis_funcs=10,
    decay_rates=np.arange(1, 10 + 1),
    enforce_decay_to_zero=True,
    order=4,
    width=2,
    frequencies=3,
)


def plot_basis(cls):
    cls_params = cls._get_param_names()
    new_kwargs = trim_kwargs(cls, KWARGS.copy(), {cls.__name__: cls_params})
    bas = cls(**new_kwargs)
    fig, ax = plt.subplots(1, 1, figsize=(5, 2.5))
    ax.plot(*bas.evaluate_on_grid(300), lw=4)
    for side in ["left", "right", "top", "bottom"]:
        ax.spines[side].set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])
    plt.subplots_adjust(left=0.02, right=0.98, top=0.98, bottom=0.02)


def plot_basis_from_values(x, bx):
    fig, ax = plt.subplots(1, 1, figsize=(5, 2.5))
    ax.plot(x, bx, lw=4)
    for side in ["left", "right", "top", "bottom"]:
        ax.spines[side].set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])
    plt.subplots_adjust(left=0.02, right=0.98, top=0.98, bottom=0.02)


def plot_raised_cosine_linear():
    plot_basis(nmo.basis.RaisedCosineLinearEval)


def plot_raised_cosine_log():
    plot_basis(nmo.basis.RaisedCosineLogEval)


def plot_mspline():
    plot_basis(nmo.basis.MSplineEval)


def plot_bspline():
    plot_basis(nmo.basis.BSplineEval)


def plot_cyclic_bspline():
    plot_basis(nmo.basis.CyclicBSplineEval)


def plot_orth_exp_basis():
    plot_basis(nmo.basis.OrthExponentialEval)


def plot_fourier_basis():
    plot_basis(nmo.basis.FourierEval)


def plot_nd_basis_thumbnail():
    a_basis = nmo.basis.MSplineEval(n_basis_funcs=15, order=3)
    b_basis = nmo.basis.RaisedCosineLogEval(n_basis_funcs=14)
    prod_basis = a_basis * b_basis

    x_coord = np.linspace(0, 1, 1000)
    y_coord = np.linspace(0, 1, 1000)

    X, Y, Z = prod_basis.evaluate_on_grid(200, 200)

    # basis element pairs
    element_pairs = [[0, 0], [5, 1], [10, 5]]

    # plot the 1D basis element and their product
    fig, axs = plt.subplots(3, 3, figsize=(8, 6))
    cc = 0
    for i, j in element_pairs:
        # plot the element form a_basis
        axs[cc, 0].plot(x_coord, a_basis.compute_features(x_coord), "grey", alpha=0.3)
        axs[cc, 0].plot(x_coord, a_basis.compute_features(x_coord)[:, i], "b")
        axs[cc, 0].set_title(f"$a_{{{i}}}(x)$", color="b")

        # plot the element form b_basis
        axs[cc, 1].plot(y_coord, b_basis.compute_features(y_coord), "grey", alpha=0.3)
        axs[cc, 1].plot(y_coord, b_basis.compute_features(y_coord)[:, j], "b")
        axs[cc, 1].set_title(f"$b_{{{j}}}(y)$", color="b")

        # select & plot the corresponding product basis element
        k = i * b_basis.n_basis_funcs + j
        axs[cc, 2].contourf(X, Y, Z[:, :, k], cmap="Blues")
        axs[cc, 2].set_title(
            rf"$A_{{{k}}}(x,y) = a_{{{i}}}(x) \cdot b_{{{j}}}(y)$", color="b"
        )
        axs[cc, 2].set_xlabel("x-coord")
        axs[cc, 2].set_ylabel("y-coord")
        axs[cc, 2].set_aspect("equal")

        cc += 1
    axs[2, 0].set_xlabel("x-coord")
    axs[2, 1].set_xlabel("y-coord")

    plt.tight_layout()


def plot_1d_basis_thumbnail():
    order = 4
    n_basis = 10
    bspline = nmo.basis.BSplineEval(n_basis_funcs=n_basis, order=order)
    x, y = bspline.evaluate_on_grid(100)

    plt.figure(figsize=(5, 3))
    plt.plot(x, y, lw=2)
    plt.title("B-Spline Basis")


def plot_identity_basis():
    plt.figure()
    ax = plt.subplot(111, aspect="equal")
    plt.plot([-1, 1], [-1, 1], lw=6, color="tomato", label="f(x) = x")
    ax.axhline(0, -1, 1, color="k")
    ax.axvline(0, -1, 1, color="k")
    for side in ["left", "right", "top", "bottom"]:
        ax.spines[side].set_visible(False)
    plt.xticks([])
    plt.yticks([])
    plt.tight_layout()
    plt.legend(fontsize=30)
    plt.show()


def plot_history_basis():
    n_basis = KWARGS["n_basis_funcs"]
    fig, ax = plt.subplots(1, 1, figsize=(5, 2.5))
    for i in np.linspace(0, 1, n_basis):
        plt.plot([i, i], [0, 1], lw=4)
    for side in ["right", "top"]:
        ax.spines[side].set_visible(False)
    plt.gca().spines["bottom"].set_linewidth(2)
    plt.gca().spines["left"].set_linewidth(2)
    ax.set_xticks([])
    ax.set_yticks([])
    plt.ylim(0, 1.2)
    plt.xlabel("lag", fontsize=20)
    plt.subplots_adjust(left=0.02, right=0.98, top=0.98, bottom=0.15)


def plot_laguerre_basis():

    def laguerre_poly(x, poly_coef, decay_rate):
        return jnp.exp(-decay_rate * x / 2) * jnp.polyval(
            poly_coef[::-1], decay_rate * x
        )

    x = jnp.linspace(0, 30, 1000)
    c = 1.0
    N = 5
    P = np.zeros((N, N))
    for n in range(N):
        P[n, : (n + 1)] = laguerre(n).coef[::-1]
    P = jnp.array(P)
    vmap_laguerre = jax.vmap(
        lambda x, p: laguerre_poly(x, p, c), in_axes=(None, 0), out_axes=1
    )
    plot_basis_from_values(x, vmap_laguerre(x, P))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bactracking line search algorithm."""

from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import Optional
from typing import Union

from dataclasses import dataclass

import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src.cond import cond
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_conj
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot_real


class BacktrackingLineSearchState(NamedTuple):
  """Named tuple containing state information."""
  iter_num: int
  params: Any
  value: float
  grad: Any  # either initial or final for armijo or glodstein
  value_init: float 
  grad_init: Any
  error: float
  done: bool
  failed: bool
  num_fun_eval: int
  num_grad_eval: int
  aux: Optional[Any] = None


@dataclass(eq=False)
class BacktrackingLineSearch(base.IterativeLineSearch):
  """Backtracking line search.

  Supports complex variables.

  Attributes:
    fun: a function of the form ``fun(params, *args, **kwargs)``, where
      ``params`` are parameters of the model,
      ``*args`` and ``**kwargs`` are additional arguments.
    value_and_grad: if ``False``, ``fun`` should return the function value only.
      If ``True``, ``fun`` should return both the function value and the
      gradient. If it is a Callable, fun should return the value while value_and_grad
      returns value and gradient of the objective.
    has_aux: if ``False``, ``fun`` should return the function value only.
      If ``True``, ``fun`` should return a pair ``(value, aux)`` where ``aux``
      is a pytree of auxiliary values.

    condition: either "armijo", "goldstein", "strong-wolfe" or "wolfe".
    c1: constant used by the (strong) Wolfe condition.
    c2: constant strictly less than 1 used by the (strong) Wolfe condition.
    decrease_factor: factor by which to decrease the stepsize during line search
      (default: 0.8).
    max_stepsize: upper bound on stepsize.

    maxiter: maximum number of line search iterations.
    tol: tolerance of the stopping criterion.

    verbose: whether to print information on every iteration or not.

    jit: whether to JIT-compile the optimization loop (default: "auto").
    unroll: whether to unroll the optimization loop (default: "auto").
  """
  fun: Callable
  value_and_grad: Union[bool, Callable] = False
  has_aux: bool = False

  maxiter: int = 30
  tol: float = 0.0

  condition: str = "strong-wolfe"
  c1: float = 1e-4
  c2: float = 0.9
  decrease_factor: float = 0.8
  max_stepsize: float = 1.0

  verbose: Union[bool, int] = False
  jit: base.AutoOrBoolean = "auto"
  unroll: base.AutoOrBoolean = "auto"

  def init_state(
      self,
      init_stepsize: float,
      params: Any,
      value: Optional[float] = None,
      grad: Optional[Any] = None,
      descent_direction: Optional[Any] = None,
      fun_args: list = [],
      fun_kwargs: dict = {},
  ) -> BacktrackingLineSearchState:
    """Initialize the line search state.

    Args:
      init_stepsize: initial step size value.
      params: current parameters.
      value: current function value (recomputed if None).
      grad: current gradient (recomputed if None).
      descent_direction: ignored.
      fun_args: additional positional arguments to be passed to ``fun``.
      fun_kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      state
    """
    del descent_direction  # Not used.
    del init_stepsize  # Not used.

    num_fun_eval = jnp.asarray(0, base.NUM_EVAL_DTYPE)
    num_grad_eval = jnp.asarray(0, base.NUM_EVAL_DTYPE)
    if value is None or grad is None:
      (value, _), grad = self._value_and_grad_fun_with_aux(
          params, *fun_args, **fun_kwargs
      )
      num_fun_eval += 1
      num_grad_eval += 1

    return BacktrackingLineSearchState(iter_num=jnp.asarray(0),
                                       params=params,
                                       value=value,
                                       grad=grad,
                                       value_init=value,
                                       grad_init=grad,
                                       aux=None,  # we do not need to have aux
                                       # in the initial state
                                       error=jnp.asarray(jnp.inf),
                                       done=jnp.asarray(False),
                                       failed=jnp.asarray(False),
                                       num_fun_eval=num_fun_eval,
                                       num_grad_eval=num_grad_eval)

  def update(
      self,
      stepsize: float,
      state: NamedTuple,
      params: Any,
      value: Optional[float] = None,
      grad: Optional[Any] = None,
      descent_direction: Optional[Any] = None,
      fun_args: list = [],
      fun_kwargs: dict = {},
  ) -> base.LineSearchStep:
    """Performs one iteration of backtracking line search.

    Args:
      stepsize: current estimate of the step size.
      state: named tuple containing the line search state.
      params: current parameters.
      value: current function value (computed at initialization, unused here).
      grad: current gradient (computed at initialization, unused here).
      descent_direction: descent direction (negative gradient if None).
      fun_args: additional positional arguments to be passed to ``fun``.
      fun_kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      (params, state)
    """
    # Ensure that stepsize does not exceed upper bound.
    stepsize = jnp.minimum(self.max_stepsize, stepsize)
    num_fun_eval = state.num_fun_eval
    num_grad_eval = state.num_grad_eval

    # Grab value and grad from initialization and avoid recomputing them
    del value
    del grad
    value = state.value_init
    grad = state.grad_init

    if descent_direction is None:
      descent_direction = tree_scalar_mul(-1, tree_conj(grad))

    slope = tree_vdot_real(tree_conj(grad), descent_direction)

    # For backtracking linesearches, we want to compute the next point
    # from the basepoint. i.e. x_i  = x_0 + s_i * p
    new_params = tree_add_scalar_mul(params, stepsize, descent_direction)
    # Every condition requires the new function value, but not every one
    # requires the new gradient value (we'll assume that this code is called
    # under `jit`).
    num_fun_eval += 1
    if self.condition in ["armijo", "goldstein"]:
      new_value, new_aux = self._fun_with_aux(
          new_params, *fun_args, **fun_kwargs
      )
      # For those conditions, no need to compute a new grad
      # We recompute a new_grad only once we have found the right stepsize,
      # see below
      new_grad, new_slope = grad, slope
    else:
      (new_value, new_aux), new_grad = self._value_and_grad_fun_with_aux(
          new_params, *fun_args, **fun_kwargs
      )
      new_slope = tree_vdot_real(tree_conj(new_grad), descent_direction)
      num_grad_eval += 1

    # Armijo condition (upper bound on admissible step size).
    # cond1 = new_value <= value + self.c1 * stepsize * slope
    # See equation (3.6a), Numerical Optimization, Second edition.
    diff_cond1 = new_value - (value + self.c1 * stepsize * slope)
    error_cond1 = jnp.where(jnp.isnan(diff_cond1), jnp.inf, diff_cond1)
    error_cond1 = jnp.maximum(error_cond1, 0.0)
    error = error_cond1

    if self.condition == "armijo":
      # We don't need to do any extra work, since this is covered by the
      # sufficient decrease condition above.
      pass

    elif self.condition == "strong-wolfe":
      # cond2 = abs(new_slope) <= c2 * abs(slope)
      # See equation (3.7b), Numerical Optimization, Second edition.
      diff_cond2 = jnp.abs(new_slope) - self.c2 * jnp.abs(slope)
      error_cond2 = jnp.where(jnp.isnan(diff_cond2), jnp.inf, diff_cond2)
      error_cond2 = jnp.maximum(error_cond2, 0.0)
      error = jnp.maximum(error_cond1, error_cond2)

    elif self.condition == "wolfe":
      # cond2 = new_slope >= c2 * slope
      # See equation (3.6b), Numerical Optimization, Second edition.
      diff_cond2 = self.c2 * slope - new_slope
      error_cond2 = jnp.where(jnp.isnan(diff_cond2), jnp.inf, diff_cond2)
      error_cond2 = jnp.maximum(error_cond2, 0.0)
      error = jnp.maximum(error_cond1, error_cond2)

    elif self.condition == "goldstein":
      # cond2 = new_value >= value + (1 - self.c1) * stepsize * slope
      diff_cond2 = value + (1 - self.c1) * stepsize * slope - new_value
      error_cond2 = jnp.where(jnp.isnan(diff_cond2), jnp.inf, diff_cond2)
      error_cond2 = jnp.maximum(error_cond2, 0.0)
      error = jnp.maximum(error_cond1, error_cond2)

    else:
      raise ValueError("condition should be one of "
                       "'armijo', 'goldstein', 'strong-wolfe' or 'wolfe'.")

    done = state.done | (error <= self.tol)
    failed = state.failed | ((state.iter_num + 1 == self.maxiter) & ~done)

    new_stepsize = jnp.where(done | failed,
                             stepsize,
                             stepsize * self.decrease_factor)

    if self.condition in ["armijo", "goldstein"]:
      # If we are done for the armijo or the goldstein conditions,
      # we compute the final gradient (we had not computed it before since 
      # these conditions did not require it)
      new_grad = cond(done | failed,
                      self._compute_final_grad,
                      lambda *_: grad, 
                      new_params, fun_args, fun_kwargs,
                      jit=self.jit)
      maybe_additional_eval = jnp.asarray(done | failed, dtype=base.NUM_EVAL_DTYPE)
      num_grad_eval = num_grad_eval + maybe_additional_eval
      # We a priori always access the function value when computing the gradient
      num_fun_eval = num_fun_eval + maybe_additional_eval

    new_state = state._replace(iter_num=state.iter_num + 1,
                               params=new_params,
                               value=new_value,
                               grad=new_grad,
                               aux=new_aux,
                               done=done,
                               error=error,
                               failed=failed,
                               num_fun_eval=num_fun_eval,
                               num_grad_eval=num_grad_eval)

    if self.verbose:
      additional_info = {'Stepsize': stepsize, 'Objective Value': new_value}
      if self.condition != 'armijo':
        error_name = "Minimum Decrease & Curvature Errors"
        additional_info.update({'Decrease Error': error_cond1})
      else:
        error_name = "Decrease Error"
      self.log_info(
          new_state,
          error_name=error_name,
          additional_info=additional_info
      )
    return base.LineSearchStep(stepsize=new_stepsize, state=new_state)
  
  def _compute_final_grad(self, params, fun_args, fun_kwargs):
    return self._grad_with_aux(params, *fun_args, **fun_kwargs)[0]

  def __post_init__(self):
    self._fun_with_aux, self._grad_with_aux, self._value_and_grad_fun_with_aux = \
        base._make_funs_with_aux(fun=self.fun,
                                 value_and_grad=self.value_and_grad,
                                 has_aux=self.has_aux)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional, Tuple, Union

import jax

from ..observation_models import PoissonObservations
from ..regularizer import Ridge
from ._svrg_defaults import (
    glm_softplus_poisson_l_max_and_l,
    svrg_optimal_batch_and_stepsize,
)

if TYPE_CHECKING:
    from ..glm import GLM, PopulationGLM


def glm_compute_optimal_stepsize_configs(
    model: Union[GLM, PopulationGLM],
) -> Tuple[Optional[Callable], Optional[Callable], Optional[float]]:
    """
    Compute configuration functions for optimal step size selection based on the model.

    This function returns a tuple of three elements that are used for configuring the
    optimal step size and batch size for variance reduced gradient (SVRG and
    ProxSVRG) algorithms. If the model is configured with specific solver names,
    the appropriate computation functions are returned. Additionally, it determines the
    smoothness and strong convexity constants based on the model's observation and regularizer.

    Parameters
    ----------
    model :
        The generalized linear model object for which the optimal step size and batch
        configuration need to be computed.

    Returns
    -------
    compute_optimal_params :
        A function to compute the optimal batch size and step size if the model
        is configured with the SVRG or ProxSVRG solver, None otherwise.

    compute_smoothness :
        A function to compute the smoothness constant of the loss function if the
        observation model uses a softplus inverse link function and is a Poisson
        observation model, None otherwise.

    strong_convexity :
        The strong convexity constant of the loss function if the model has a
        Ridge regularizer. If the model does not have a Ridge regularizer, this
        value will be None.

    """
    # initialize funcs and strong convexity constant
    compute_optimal_params = None
    compute_smoothness = None
    strong_convexity = (
        None if not isinstance(model.regularizer, Ridge) else model.regularizer_strength
    )

    # look-up table for selecting the optimal step and batch
    if model.solver_name in ("SVRG", "ProxSVRG"):
        compute_optimal_params = svrg_optimal_batch_and_stepsize

    # get the smoothness parameter compute function
    if model.inverse_link_function is jax.nn.softplus and isinstance(
        model.observation_model, PoissonObservations
    ):
        compute_smoothness = glm_softplus_poisson_l_max_and_l

    return compute_optimal_params, compute_smoothness, strong_convexity



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp
from jax.test_util import check_grads
from jax.tree_util import tree_map
import numpy as onp

from sklearn import datasets
from sklearn import preprocessing
from sklearn import svm

from nemos.third_party.jaxopt.jaxopt import projection
from nemos.third_party.jaxopt.jaxopt._src.base import KKTSolution
from nemos.third_party.jaxopt.jaxopt._src.osqp import BoxOSQP
from nemos.third_party.jaxopt.jaxopt._src.osqp import OSQP
from nemos.third_party.jaxopt.jaxopt._src.osqp import extract_Qc_from_obj
from nemos.third_party.jaxopt.jaxopt._src.cvxpy_wrapper import CvxpyQP
from nemos.third_party.jaxopt.jaxopt._src import test_util


def get_random_osqp_problem(problem_size, eq_constraints, ineq_constraints):
  assert problem_size >= eq_constraints  # very likely to be infeasible
  onp.random.seed(problem_size + eq_constraints + ineq_constraints)
  Q = onp.random.randn(problem_size, problem_size)
  Q = Q.T.dot(Q)  # PSD matrix
  c = onp.random.randn(problem_size)
  A = onp.random.randn(ineq_constraints + eq_constraints, problem_size)
  l = onp.random.randn(ineq_constraints)
  u = l + jnp.abs(onp.random.randn(ineq_constraints))  # l < u
  b = onp.random.randn(eq_constraints)  # Ax = b
  l = jnp.concatenate([l, b])
  u = jnp.concatenate([u, b])
  return (Q, c), A, (l, u)


def _from_osqp_form_to_boxosqp_form(Q, c, A, l, u):
  """Return form used by CvxpyQP from BoxOSQP form (support only matrices)."""
  is_eq = l == u
  is_ineq_l = jnp.logical_and(l != u, l != -jnp.inf)
  is_ineq_u = jnp.logical_and(l != u, u != jnp.inf)
  if jnp.any(is_eq):
    A_eq, b = A[is_eq,:], l[is_eq]
  else:
    A_eq, b = jnp.zeros((1,len(c))), jnp.zeros((1,))
  if jnp.any(is_ineq_l) and jnp.any(is_ineq_u):
    G = jnp.concatenate([-A[is_ineq_l,:],A[is_ineq_u,:]])
    h = jnp.concatenate([-l[is_ineq_l],u[is_ineq_u]])
  elif jnp.any(is_ineq_l):
    G, h = -A[is_ineq_l,:], -l[is_ineq_l]
  elif jnp.any(is_ineq_u):
    G, h = A[is_ineq_u,:], u[is_ineq_u]
  else:
    return (Q, c), (A_eq, b), None
  return (Q, c), (A_eq, b), (G, h)


class BoxOSQPTest(test_util.JaxoptTestCase):

  def test_small_qp(self):
    # Setup a random QP min_x 0.5*x'*Q*x + q'*x s.t. Ax = z; l <= z <= u;
    eq_constraints, ineq_constraints = 2, 4
    onp.random.seed(42)
    problem_size = 16
    params_obj, params_eq, params_ineq = get_random_osqp_problem(problem_size, eq_constraints, ineq_constraints)
    tol = 1e-5
    osqp = BoxOSQP(tol=tol)
    params, state = osqp.run(None, params_obj, params_eq, params_ineq)
    self.assertLessEqual(state.error, tol)
    opt_error = osqp.l2_optimality_error(params, params_obj, params_eq, params_ineq)
    self.assertAllClose(opt_error, 0.0, atol=1e-4)

    def test_against_cvxpy(params_obj):
      (Q, c), Ab, Gh = _from_osqp_form_to_boxosqp_form(params_obj[0], params_obj[1],
                                                       params_eq, params_ineq[0], params_ineq[1])
      Q = 0.5 * (Q + Q.T)
      qp = CvxpyQP()
      hyperparams = dict(params_obj=(Q, c), params_eq=Ab, params_ineq=Gh)
      sol = qp.run(None, **hyperparams).params
      return sol.primal

    atol = 1e-4
    cvx_primal = test_against_cvxpy(params_obj)
    self.assertArraysAllClose(params.primal[0], cvx_primal, atol=atol)

    def osqp_run(params_obj):
      Q, c = params_obj
      Q = 0.5 * (Q + Q.T)
      sol = osqp.run(None, (Q, c), params_eq, params_ineq).params
      return sol.primal[0]

    jacosqp = jax.jacrev(osqp_run)(params_obj)
    jaccvxpy = jax.jacrev(test_against_cvxpy)(params_obj)
    self.assertArraysAllClose(jacosqp[0], jaccvxpy[0], atol=5e-2)
    self.assertArraysAllClose(jacosqp[1], jaccvxpy[1], atol=5e-2)

  @parameterized.product(derivative_with_respect_to=["eq", "ineq"])  # "none", "obj" are covered partially in other tests.
  def test_qp_eq_and_ineq(self, derivative_with_respect_to):
    Q = 2 * jnp.array([[2.0, 0.5], [0.5, 1]])
    c = jnp.array([1.0, 1.0])
    A_eq = jnp.array([[1.0, 1.0]])
    b = jnp.array([1.0])
    G = jnp.array([[-1.0, 0.0], [0.0, -1.0]])
    h = jnp.array([0.0, 0.0])
    tol = 1e-5
    osqp = BoxOSQP(tol=tol, verbose=0)

    @jax.jit
    def osqp_run(Q, c, A_eq, G, b, h):
      l = jnp.concatenate([b, jnp.full(h.shape, -jnp.inf)])
      u = jnp.concatenate([b, jnp.full(h.shape, h)])
      Q = 0.5 * (Q + Q.T) # we need to ensure that P is symmetric even after directional perturbations
      A = jnp.concatenate([A_eq, G], axis=0)
      return osqp.run(None, (Q, c), A, (l, u))

    atol = 1e-2
    eps = 1e-4

    if "none" == derivative_with_respect_to:
      params, state = osqp_run(Q, c, A_eq, G, b, h)
      self.assertLessEqual(state.error, tol)
      assert state.status == BoxOSQP.SOLVED

      qp = CvxpyQP()
      hyperparams_qp = dict(params_obj=(Q, c), params_eq=(A_eq, b), params_ineq=(G, h))
      params_qp, state = qp.run(None, **hyperparams_qp)
      self.assertArraysAllClose(params.primal[0], params_qp.primal, atol=atol)

    def keep_ineq_only(params):
      b_idx = int(b.shape[0])
      mu, phi = params.dual_ineq
      return KKTSolution(params.primal, params.dual_eq, (mu[b_idx:], phi[b_idx:]))

    if "obj" == derivative_with_respect_to:
      solve_run_c = lambda c: keep_ineq_only(osqp_run(Q, c, A_eq, G, b, h).params)
      check_grads(solve_run_c, args=(c,), order=1, modes=['rev'], eps=eps, atol=atol)
      solve_run_Q = lambda Q: keep_ineq_only(osqp_run(Q, c, A_eq, G, b, h).params)
      check_grads(solve_run_Q, args=(Q,), order=1, modes=['rev'], eps=eps, atol=atol)

    if "eq" == derivative_with_respect_to:
      solve_run_A_eq = lambda A_eq: keep_ineq_only(osqp_run(Q, c, A_eq, G, b, h).params)
      check_grads(solve_run_A_eq, args=(A_eq,), order=1, modes=['rev'], eps=eps, atol=atol)
      solve_run_b = lambda b: keep_ineq_only(osqp_run(Q, c, A_eq, G, b, h).params)
      check_grads(solve_run_b, args=(b,), order=1, modes=['rev'], eps=eps, atol=atol)

    if "ineq" == derivative_with_respect_to:
      solve_run_G = lambda G: keep_ineq_only(osqp_run(Q, c, A_eq, G, b, h).params)
      check_grads(solve_run_G, args=(G,), order=1, modes=['rev'], eps=eps, atol=atol)
      solve_run_h = lambda h: keep_ineq_only(osqp_run(Q, c, A_eq, G, b, h).params)
      check_grads(solve_run_h, args=(h,), order=1, modes=['rev'], eps=eps, atol=atol)

  @parameterized.product(implicit_diff=[True, False])
  def test_projection_hyperplane(self, implicit_diff):
    x = jnp.array([1.0, 2.0])
    a = jnp.array([-0.5, 1.5])
    b = 0.3
    q = -x
    # Find ||y-x||^2 such that jnp.dot(y, a) = b.

    matvec_Q = lambda params_Q,u: u
    matvec_A = lambda params_A,u: jnp.dot(a, u).reshape(1)

    tol = 1e-4
    osqp = BoxOSQP(matvec_Q=matvec_Q, matvec_A=matvec_A, tol=tol, verbose=0)
    sol, state = osqp.run(None, (None, q), None, (b, b))

    assert state.status == BoxOSQP.SOLVED
    self.assertLessEqual(state.error, tol)
    atol = 1e-3
    opt_error = osqp.l2_optimality_error(sol, (None, q), None, (b, b))
    self.assertAllClose(opt_error, 0.0, atol=atol)
    expected = projection.projection_hyperplane(x, (a, b))
    self.assertArraysAllClose(sol.primal[0], expected, atol=atol)

  def test_projection_simplex(self):
    @jax.jit
    def _projection_simplex_qp(x, s=1.0):
      Q = jnp.eye(len(x))
      A_eq = jnp.array([jnp.ones_like(x)])
      b = jnp.array([s])
      G = -jnp.eye(len(x))
      h = jnp.zeros_like(x)
      A = jnp.concatenate([A_eq, G])
      l = jnp.concatenate([b, jnp.full(h.shape, -jnp.inf)])
      u = jnp.concatenate([b, h])
      hyperparams = dict(params_obj=(Q, -x), params_eq=A,
                         params_ineq=(l, u))

      osqp = BoxOSQP(tol=1e-5)
      return osqp.run(None, **hyperparams).params.primal[0]

    atol = 1e-3

    rng = onp.random.RandomState(0)
    x = jnp.array(rng.randn(10))
    p = projection.projection_simplex(x)
    p2 = _projection_simplex_qp(x)
    self.assertArraysAllClose(p, p2, atol=atol)

    J = jax.jacrev(projection.projection_simplex)(x)
    J2 = jax.jacrev(_projection_simplex_qp)(x)
    self.assertArraysAllClose(J, J2, atol=atol)

  def test_eq_constrained_qp_with_pytrees(self):
    rng = onp.random.RandomState(0)
    Q = rng.randn(7, 7)
    Q = onp.dot(Q, Q.T)
    A = rng.randn(4, 7)

    tmp = rng.randn(7)
    # Must have the same pytree structure as the output of matvec_P.
    c = {'foo':tmp[:3], 'bar':tmp[3:]}
    # Must have the same pytree structure as the output of matvec_A.
    b = [[rng.randn(1)] for _ in range(4)]

    def matvec_Q(Q, dic):
      x_ = jnp.concatenate([dic['foo'], dic['bar']])
      res = jnp.dot(Q, x_)
      return {'foo':res[:3], 'bar':res[3:]}

    def matvec_A(A, dic):
      x_ = jnp.concatenate([dic['foo'], dic['bar']])
      z = jnp.dot(A, x_)
      ineqs = jnp.split(z,z.shape[0])
      return [[ineq] for ineq in ineqs]

    tol = 1e-5
    atol = 1e-4

    # With pytrees directly.
    hyperparams = dict(params_obj=(Q, c), params_eq=A, params_ineq=(b, b))
    osqp = BoxOSQP(matvec_Q=matvec_Q, matvec_A=matvec_A, tol=tol)
    # sol.primal has the same pytree structure as the output of matvec_Q.
    # sol.dual_eq has the same pytree structure as the output of matvec_A.
    sol_pytree, state = osqp.run(None, **hyperparams)
    assert state.status == BoxOSQP.SOLVED
    self.assertAllClose(osqp.l2_optimality_error(sol_pytree, **hyperparams), 0.0, atol=atol)

    flat_x = lambda x: jnp.concatenate([x['foo'], x['bar']])
    flat_z = lambda z: jnp.concatenate([zi[0] for zi in z])

    # With flattened pytrees.
    c_flat = flat_x(c)
    b_flat = flat_z(b)
    hyperparams = dict(params_obj=(Q, c_flat), params_eq=A, params_ineq=(b_flat, b_flat))
    osqp = BoxOSQP(tol=tol)
    sol = osqp.run(None, **hyperparams).params
    self.assertAllClose(osqp.l2_optimality_error(sol, **hyperparams), 0.0, atol=atol)

    # Check that the solutions match.
    self.assertArraysAllClose(flat_x(sol_pytree.primal[0]), sol.primal[0], atol=atol)
    self.assertArraysAllClose(flat_z(sol_pytree.primal[1]), sol.primal[1], atol=atol)
    self.assertArraysAllClose(flat_z(sol_pytree.dual_eq), sol.dual_eq, atol=atol)

  def test_binary_kernel_svm(self):
    n_samples, n_features = 50, 8
    n_informative = (3*n_features) // 4
    # Prepare data.
    X, y = datasets.make_classification(n_samples=n_samples, n_features=n_features,
                                        n_informative=n_informative, n_classes=2,
                                        random_state=0)
    X = preprocessing.Normalizer().fit_transform(X)
    y = y * 2 - 1.  # Transform labels from {0, 1} to {-1, 1}.
    lam = 10.0
    C = 1./ lam

    K = jnp.dot(X, X.T)

    # The dual objective is:
    # fun(beta) = 0.5 beta^T K beta - beta^T y
    # subject to
    # sum(beta) = 0
    # 0 <= beta_i <= C if y_i = 1
    # -C <= beta_i <= 0 if y_i = -1
    # where C = 1.0 / lam
    matvec_A = lambda _,b: (b, jnp.sum(b))
    l = -jax.nn.relu(-y * C), 0.
    u =  jax.nn.relu( y * C), 0.
    hyper_params = dict(params_obj=(K, -y), params_eq=None, params_ineq=(l, u))
    tol = 1e-5
    osqp = BoxOSQP(matvec_A=matvec_A, tol=tol)

    sol, state = osqp.run(None, **hyper_params)
    self.assertLessEqual(state.error, tol)
    atol = 1e-3
    opt_error = osqp.l2_optimality_error(sol, **hyper_params)
    self.assertAllClose(opt_error, 0.0, atol=atol)

    def binary_kernel_svm_skl(K, y):
      svc = svm.SVC(kernel="precomputed", C=C, tol=tol).fit(K, y)
      dual_coef = onp.zeros(K.shape[0])
      dual_coef[svc.support_] = svc.dual_coef_[0]
      return dual_coef

    beta_fit_skl = binary_kernel_svm_skl(K, y)
    # we solve the dual problem with BoxOSQP so the dual of svm.SVC
    # corresponds to the primal variables of BoxOSQP solution.
    self.assertAllClose(sol.primal[0], beta_fit_skl, atol=1e-2)

  def test_infeasible_polyhedron(self):
    # argmin_p \|p - x\|_2 = argmin_p <p,Ip> - 2<x,p> = argmin_p 0.5pIp - <x,p>
    # under p1 + p2 =  1
    #            p2 =  1
    #      -p1 + p2 = -1
    #      -p1     <=  0
    #      -p2     <=  0
    # This QP is primal/dual infeasible.
    A = jnp.array([[1.0, 1.0],[0.0, 1.0],[-1.0, 1.0]])
    b = jnp.array([1.0, 1.0, -1.0])
    G = jnp.array([[-1.0, 0.0], [0.0, -1.0]])
    h = jnp.array([0.0, 0.0])

    l = b, jnp.full(h.shape, -jnp.inf)
    u = b, h

    def matvec_A(_, p):
      return jnp.dot(A, p), jnp.dot(G, p)

    x = jnp.zeros(2)
    I = jnp.eye(len(x))
    hyper_params = dict(params_obj=(I, -x), params_eq=None, params_ineq=(l, u))
    osqp = BoxOSQP(matvec_A=matvec_A, check_primal_dual_infeasability=True, tol=1e-5)
    sol, state = osqp.run(None, **hyper_params)
    assert state.status in [BoxOSQP.PRIMAL_INFEASIBLE, BoxOSQP.DUAL_INFEASIBLE]

  def test_infeasible_primal_only(self):
    # argmin   x1 + x2
    # under    x1 >= 6
    #          x2 >= 6
    #     x1 + x2 <= 11
    # This QP is primal infeasible.
    Q = jnp.zeros((2,2))
    c = jnp.array([1.,1.])

    def matvec_A(_, x):
      return x[0], x[1], x[0] + x[1]
    l = 6., 6., -jnp.inf
    u = jnp.inf, jnp.inf, 11.

    hyper_params = dict(params_obj=(Q, c), params_eq=None, params_ineq=(l, u))
    osqp = BoxOSQP(matvec_A=matvec_A, check_primal_dual_infeasability=True)
    sol, state = osqp.run(None, **hyper_params)
    assert state.status == BoxOSQP.PRIMAL_INFEASIBLE

  def test_unbounded_primal(self):
    # argmin   x1 -2x2 + x3
    # under  x1 + x2 >= 0
    #              x3 = 1
    # This (degenerated) QP is dual infeasible (unbounded primal).
    def matvec_A(_, x):
      return x[0] + x[1], x[2]
    l = 0., 1.
    u = jnp.inf, 1.

    def matvec_Q(_, x):
      return 0., 0., 0.

    hyper_params = dict(params_obj=(None, (-1., -2., 1.)), params_eq=None, params_ineq=(l, u))
    osqp = BoxOSQP(matvec_Q=matvec_Q, matvec_A=matvec_A,
                check_primal_dual_infeasability=True,
                tol=1e-3)
    init_params = osqp.init_params(init_x=(5., -3., 0.02), **hyper_params)
    sol, state = osqp.run(init_params, **hyper_params)
    assert state.status == BoxOSQP.DUAL_INFEASIBLE

  def test_jacobi_preconditioner(self):
    # Portfolio optimization
    # Problems of the form
    #     max_x mu^Tx - gamma x Cov x
    #     under sum_i x_i  = 1
    #                 x_i >= 0
    # with Cov = F@F.T + D the risk model, F @ F.T low rank, and D diagonal
    # mu the expected return
    onp.random.seed(0)
    assets, factors = 10, 3
    gamma = 1
    F = onp.random.randn(assets, factors)
    D = jnp.diag(onp.random.uniform(0., factors ** 0.5, assets))
    Cov = 2*gamma*(F @ F.T + D)   # minimize risk
    mu = onp.random.randn(assets)   # maximize gain

    A = jnp.concatenate([jnp.ones((1, assets)), jnp.eye(assets)], axis=0)
    l = jnp.concatenate([jnp.array([1.]), jnp.zeros(assets)])
    u = jnp.concatenate([jnp.array([1.]), jnp.full(assets, jnp.inf)])

    tol = 1e-5
    atol = 1e-3

    hyper_params = dict(params_obj=(Cov, mu), params_eq=A, params_ineq=(l, u))
    osqp = BoxOSQP(tol=tol, eq_qp_solve='cg+jacobi')
    sol, state = osqp.run(None, **hyper_params)
    self.assertLessEqual(state.error, tol)
    assert state.status == BoxOSQP.SOLVED
    opt_error = osqp.l2_optimality_error(sol, **hyper_params)
    self.assertAllClose(opt_error, 0.0, atol=atol)

    # Default preconditioner: Identity
    # Jacobi preconditioner J should be better.
    # We can check this heuristically with:
    #    || M@P@x - x ||_2 where preconditioner P approximates M^{-1}
    # Error for Default preconditioner:    M@x   - x
    # Error for Jacobi preconditioner:     M@J@x - x
    # The error of Jacobi is expected to be lower.
    sigma = 1e-6
    J = osqp._eq_qp_solve_impl
    x = sol.primal[0]
    for rho_bar in [1e-4, 0.001, 0.01, 0.1, 1., 10., 100., 1000, 1e4]:
      solver_state = J.init_state(x, Cov, A, sigma, rho_bar)
      M = Cov + sigma * jnp.eye(assets) + rho_bar * A.T @ A
      approx_x = M @ J._matvec_precond(solver_state, x)
      diff = jnp.sum((approx_x - x)**2)**0.5
      no_precond_approx = M @ x
      no_precond_diff = jnp.sum((no_precond_approx - x)**2)**0.5
      self.assertLessEqual(diff, no_precond_diff)

  def test_lasso(self):
    # Objective:
    #     min ||data @ x - targets||_2^2 + 2 * n * lam ||x||_1
    #
    # Converted in QP:
    #
    #     min y^Ty + 2*n*lam 1^T t
    #     under y = data @ x - targets
    #          -t <= x <= t
    #
    # With BoxOSQP formulation:
    #
    #     min y^Ty + 2*n*lam 1^T t
    #     under       targets = data @ x - y
    #           0         <= x + t <= infinity
    #           -infinity <= x - t <= 0
    #
    # Note that we use 2n corrective factor in objective to
    # optimize same objective as ProximalGradient.
    n = 10
    data, targets = datasets.make_regression(n_samples=n, n_features=3, random_state=0)
    lam = 10.0

    def matvec_Q(_, xyt):
      x, y, t = xyt
      return jnp.zeros_like(x), 2 * y, jnp.zeros_like(t)

    c = jnp.zeros(data.shape[1]), jnp.zeros(data.shape[0]), 2*n*lam * jnp.ones(data.shape[1])

    def matvec_A(data, xyt):
      x, y, t = xyt
      residuals = data @ x - y
      return residuals, x + t, x - t

    l = targets, jnp.zeros_like(c[0]), jnp.full(data.shape[1], -jnp.inf)
    u = targets, jnp.full(data.shape[1], jnp.inf), jnp.zeros_like(c[0])

    tol = 1e-4
    atol = 1e-2

    hyper_params = dict(params_obj=(None, c), params_eq=data, params_ineq=(l, u))
    osqp = BoxOSQP(matvec_Q=matvec_Q, matvec_A=matvec_A, tol=tol)
    sol, state = osqp.run(None, **hyper_params)
    self.assertLessEqual(state.error, tol)
    assert state.status == BoxOSQP.SOLVED
    opt_error = osqp.l2_optimality_error(sol, **hyper_params)
    self.assertAllClose(opt_error, 0.0, atol=atol)

    # Verify with Proximal Gradient

    def run_osqp(data):
      hyper_params = dict(params_obj=(None, c), params_eq=data, params_ineq=(l, u))
      sol, state = osqp.run(None, **hyper_params)
      return sol.primal[0][0]

    from nemos.third_party.jaxopt.jaxopt import objective
    from nemos.third_party.jaxopt.jaxopt import prox
    from nemos.third_party.jaxopt.jaxopt import ProximalGradient

    fun = objective.least_squares
    w_init = jnp.zeros(data.shape[1])
    pg = ProximalGradient(fun=fun, prox=prox.prox_lasso, maxiter=400, tol=tol,
                          acceleration=True)
    def run_prox(data):
      w_fit, info = pg.run(w_init, hyperparams_prox=lam, data=(data, targets))
      return w_fit

    w_fit = run_prox(data)
    self.assertArraysAllClose(sol.primal[0][0], w_fit, atol=atol)

    jac_osqp = jax.jacrev(run_osqp)(data)
    jac_prox = jax.jacrev(run_prox)(data)
    self.assertArraysAllClose(jac_osqp, jac_prox, atol=1e-2)

  def test_fun_api(self):
    hyper_params = dict(params_obj=None, params_eq=None, params_ineq=None)

    def fun(x, params_obj):
      del params_obj  # unused
      return 0.5*jnp.dot(x, x) + jnp.sum(x) + 17.0

    problem_size = 66
    x_init = jnp.arange(problem_size, dtype=jnp.float32)
    solver = BoxOSQP(fun=fun)
    (_, _, cste), c = extract_Qc_from_obj(x_init, hyper_params['params_obj'], solver.fun)

    self.assertAlmostEqual(cste, 17.0)
    self.assertArraysAllClose(c, jnp.ones(problem_size), atol=1e-3, rtol=1e-2)

  def test_convenience_api_random_problem(self):
    (Q, c), A , (l, u) = get_random_osqp_problem(3, 1, 1)

    params_obj = dict(foo=Q, bar=c)  # mimic arbitrary pytree.
    hyper_params = dict(params_obj=params_obj, params_eq=A, params_ineq=(l, u))
    cste = 42.
    def convex_quadratic(x, params_obj):
      Q, c = params_obj['foo'], params_obj['bar']
      return 0.5*jnp.dot(x, jnp.dot(Q, x)) + jnp.dot(c, x) + cste

    init_x = jnp.array([1., 2., 3])
    tol = 1e-4

    solver_with_fun = BoxOSQP(fun=convex_quadratic, tol=tol)
    (_, _, cste_approx), c_approx = extract_Qc_from_obj(init_x,
                                                        hyper_params['params_obj'],
                                                        solver_with_fun.fun)
    self.assertAlmostEqual(cste_approx, cste)
    self.assertArraysAllClose(c_approx, c, atol=1e-2, rtol=1e-2)

    try:
      # attempt to run without providing init_params, despite fun != None.
      init_params = solver_with_fun.init_params(init_x=None, **hyper_params)
    except ValueError as e:
      pass  # here, a failure is the expected behavior.
    else:
      self.fail("Expected ValueError when init_params is not provided (`fun` API).")

    init_params = solver_with_fun.init_params(init_x, **hyper_params)
    sol_with_fun, state_with_fun = solver_with_fun.run(init_params, **hyper_params)
    self.assertLessEqual(state_with_fun.error, tol)
    
    hyper_params_without_fun = dict(params_obj=(Q, c), params_eq=A, params_ineq=(l, u))
    solver_without_fun = BoxOSQP(tol=tol)
    init_params = solver_without_fun.init_params(init_x, **hyper_params_without_fun)
    sol_without_fun, state_without_fun = solver_without_fun.run(init_params, **hyper_params_without_fun)
    self.assertLessEqual(state_without_fun.error, tol)

    tree_map((lambda x,y: self.assertArraysAllClose(x, y, atol=1e-2)), sol_with_fun, sol_without_fun)

  def test_lu_factorization(self):
    problem_size = 200
    eq_constraints = 30
    ineq_constraints = 70
    params_obj, params_eq, params_ineq = get_random_osqp_problem(problem_size, eq_constraints, ineq_constraints)
    tol = 1e-4
    osqp = BoxOSQP(tol=tol, eq_qp_solve='lu', stepsize_updates_frequency=20)
    params, state = osqp.run(None, params_obj, params_eq, params_ineq)
    self.assertLessEqual(state.error, tol)
    opt_error = osqp.l2_optimality_error(params, params_obj, params_eq, params_ineq)
    self.assertAllClose(opt_error, 0.0, atol=1e-2)


class OSQPTest(test_util.JaxoptTestCase):

  def _check_derivative_A_b(self, solver, Q, c, A, b, params_ineq):
    @jax.jit
    def fun(Q, c, A, b):
      Q = 0.5 * (Q + Q.T)
      hyperparams = dict(params_obj=(Q, c),
                         params_eq=(A, b),
                         params_ineq=params_ineq)
      # reduce the primal variables to a scalar value for test purpose.
      return jnp.sum(solver.run(None, **hyperparams).params[0])

    # Derivative w.r.t. A.
    rng = onp.random.RandomState(0)
    V = rng.rand(*A.shape)
    V /= onp.sqrt(onp.sum(V ** 2))
    eps = 1e-3
    deriv_jax = jnp.vdot(V, jax.grad(fun, argnums=2)(Q, c, A, b))
    deriv_num = (fun(Q, c, A + eps * V, b) - fun(Q, c, A - eps * V, b)) / (2 * eps)
    self.assertAllClose(deriv_jax, deriv_num, rtol=1e-2)

    # Derivative w.r.t. b.
    v = rng.rand(*b.shape)
    v /= onp.sqrt(onp.sum(v ** 2))
    eps = 1e-3
    deriv_jax = jnp.vdot(v, jax.grad(fun, argnums=3)(Q, c, A, b))
    deriv_num = (fun(Q, c, A, b + eps * v) - fun(Q, c, A, b - eps * v)) / (2 * eps)
    self.assertAllClose(deriv_jax, deriv_num, rtol=1e-2)

  @parameterized.parameters((None, None), (jnp.dot, jnp.dot))
  def test_qp_eq_and_ineq(self, matvec_A, matvec_G):
    Q = 2 * jnp.array([[2.0, 0.5], [0.5, 1]])
    c = jnp.array([1.0, 1.0])
    A = jnp.array([[1.0, 1.0]])
    b = jnp.array([1.0])
    G = jnp.array([[-1.0, 0.0], [0.0, -1.0]])
    h = jnp.array([0.0, 0.0])
    qp = OSQP(matvec_Q=None, matvec_A=matvec_A, matvec_G=matvec_G, tol=1e-5)
    hyperparams = dict(params_obj=(Q, c), params_eq=(A, b), params_ineq=(G, h))
    sol = qp.run(None, **hyperparams).params
    self.assertAllClose(qp.l2_optimality_error(sol, **hyperparams), 0.0, atol=1e-4)
    self._check_derivative_A_b(qp, Q, c, A, b, (G, h))  # remark: slow.

  def test_projection_simplex(self):
    def _projection_simplex_qp(x, s=1.0):
      Q = jnp.eye(len(x))
      A = jnp.array([jnp.ones_like(x)])
      b = jnp.array([s])
      G = -jnp.eye(len(x))
      h = jnp.zeros_like(x)
      hyperparams = dict(params_obj=(Q, -x), params_eq=(A, b),
                         params_ineq=(G, h))

      qp = OSQP(tol=1e-5)
      # Returns the primal solution only.
      return qp.run(None, **hyperparams).params[0]

    rng = onp.random.RandomState(0)
    x = jnp.array(rng.randn(10))
    p = projection.projection_simplex(x)
    p2 = _projection_simplex_qp(x)
    self.assertArraysAllClose(p, p2, atol=1e-4)

    J = jax.jacrev(projection.projection_simplex)(x)
    J2 = jax.jacrev(_projection_simplex_qp)(x)
    self.assertArraysAllClose(J, J2, atol=1e-4)

  def test_NNLS(self):
    # Solve Non Negative Least Squares factorization.
    #
    #  min_W \|Y-UW\|_F^2
    #  s.t. W>=0
    n, m = 20, 10
    rank = 3
    onp.random.seed(654)
    U = jax.nn.relu(onp.random.randn(n, rank))
    W_0 = jax.nn.relu(onp.random.randn(m, rank))
    Y = U @ W_0.T

    def fun(W, params_obj):
      Y, U = params_obj
      return jnp.sum(jnp.square(Y - U @ W.T))

    def matvec_G(_, W):
      return -W

    zeros = jnp.zeros_like(W_0)
    hyper_params = dict(params_obj=(Y, U), params_eq=None, params_ineq=(None, zeros))

    init_W = jnp.zeros_like(W_0)
    solver = OSQP(fun=fun, matvec_G=matvec_G)
    init_params = solver.init_params(init_W, **hyper_params)
    sol, _ = solver.run(init_params=init_params, **hyper_params)
    W_sol = sol.primal

    # Check that the solution is close to the original.
    self.assertAllClose(W_0, W_sol, atol=1e-2, rtol=1e-2)

if __name__ == '__main__':
  jax.config.update("jax_enable_x64", False)
  absltest.main()



"""Base class for adapters wrapping JAXopt solvers."""

from typing import Any, Callable, ClassVar, NamedTuple, Type, TypeAlias

from nemos.third_party.jaxopt import jaxopt

from ..regularizer import Regularizer
from ._abstract_solver import OptimizationInfo, Params
from ._solver_adapter import SolverAdapter

JaxoptSolverState: TypeAlias = NamedTuple
JaxoptStepResult: TypeAlias = jaxopt.OptStep  # this is just a namedtuple(params, state)


class JaxoptAdapter(SolverAdapter[JaxoptSolverState]):
    """
    Base class for adapters wrapping JAXopt solvers.

    Besides `_solver_cls`, for proximal solvers the `_proximal` class variable
    needs to be set to `True`.
    """

    _solver_cls: ClassVar[Type]
    _proximal: ClassVar[bool] = False

    def __init__(
        self,
        unregularized_loss: Callable,
        regularizer: Regularizer,
        regularizer_strength: float | None,
        **solver_init_kwargs,
    ):
        if self._proximal:
            self.fun = unregularized_loss
            solver_init_kwargs["prox"] = regularizer.get_proximal_operator()
        else:
            self.fun = regularizer.penalized_loss(
                unregularized_loss, regularizer_strength
            )

        self.regularizer_strength = regularizer_strength

        # Prepend the regularizer strength to args for proximal solvers.
        # Methods of `jaxopt.ProximalGradient` expect `hyperparams_prox` before
        # the objective function's arguments, while others do not need this.
        self.hyperparams_prox = (self.regularizer_strength,) if self._proximal else ()

        self._solver = self._solver_cls(
            fun=self.fun,
            **solver_init_kwargs,
        )

    def init_state(self, init_params: Params, *args: Any) -> JaxoptSolverState:
        return self._solver.init_state(init_params, *self.hyperparams_prox, *args)

    def update(
        self, params: Params, state: JaxoptSolverState, *args: Any
    ) -> JaxoptStepResult:
        return self._solver.update(params, state, *self.hyperparams_prox, *args)

    def run(self, init_params: Params, *args: Any) -> JaxoptStepResult:
        return self._solver.run(init_params, *self.hyperparams_prox, *args)

    @classmethod
    def get_accepted_arguments(cls) -> set[str]:
        arguments = super().get_accepted_arguments()
        # prox is read from the regularizer, not provided as a solver argument
        if cls._proximal:
            arguments.remove("prox")
        return arguments

    def get_optim_info(self, state: JaxoptSolverState) -> OptimizationInfo:
        num_steps = state.iter_num.item()  # pyright: ignore
        function_val = (
            state.value if hasattr(state, "value") else None
        )  # pyright: ignore

        return OptimizationInfo(
            function_val=function_val,  # pyright: ignore
            num_steps=num_steps,
            converged=state.error.item() <= self.tol,  # pyright: ignore
            reached_max_steps=(num_steps == self.maxiter),
        )

    @property
    def maxiter(self):
        return self._solver.maxiter


class JaxoptProximalGradient(JaxoptAdapter):
    """
    Adapter for `jaxopt.ProximalGradient`.

    The `prox` argument passed to `jaxopt.ProximalGradient`
    is read from the regularizer.
    """

    _solver_cls = jaxopt.ProximalGradient
    _proximal = True


class JaxoptGradientDescent(JaxoptAdapter):
    """Adapter for jaxopt.GradientDescent."""

    _solver_cls = jaxopt.GradientDescent


class JaxoptBFGS(JaxoptAdapter):
    """Adapter for jaxopt.BFGS."""

    _solver_cls = jaxopt.BFGS


class JaxoptLBFGS(JaxoptAdapter):
    """Adapter for jaxopt.LBFGS."""

    _solver_cls = jaxopt.LBFGS


class JaxoptNonlinearCG(JaxoptAdapter):
    """Adapter for jaxopt.NonlinearCG."""

    _solver_cls = jaxopt.NonlinearCG



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import BFGS
from nemos.third_party.jaxopt.jaxopt import BacktrackingLineSearch
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt._src import test_util
import numpy as onp
import scipy.optimize as scipy_opt
from sklearn import datasets



N_CALLS = 0

# Uncomment this line to test in x64 
# jax.config.update('jax_enable_x64', True)

def _bfgs(fun, init, stepsize=1e-3, maxiter=500, tol=1e-3):
  value_and_grad_fun = jax.value_and_grad(fun)

  x = init
  I = jnp.eye(len(init))
  H = I
  value, grad = value_and_grad_fun(init)

  for it in range(maxiter):
    direction = -jnp.dot(H, grad)
    x_old, grad_old = x, grad
    x = x + stepsize * direction
    value, grad = value_and_grad_fun(x)

    s = x - x_old
    y = grad - grad_old
    rho = 1. / jnp.dot(s, y)

    sy = jnp.outer(s, y)
    ss = jnp.outer(s, s)
    new_H = (I - rho * sy).dot(H).dot(I - rho * sy.T) + rho * ss
    H = jnp.where(jnp.isfinite(rho), new_H, H)

    grad_norm = jnp.sqrt(jnp.sum(grad ** 2))

    if grad_norm <= tol:
      break

  return x


def get_fun(name, np):

  def rosenbrock(x):
    return np.sum(100. * np.diff(x) ** 2 + (1. - x[:-1]) ** 2)

  def himmelblau(p):
    x, y = p
    return (x ** 2 + y - 11.) ** 2 + (x + y ** 2 - 7.) ** 2

  def matyas(p):
    x, y = p
    return 0.26 * (x ** 2 + y ** 2) - 0.48 * x * y

  def eggholder(p):
    x, y = p
    return - (y + 47) * np.sin(np.sqrt(np.abs(x / 2. + y + 47.))) - x * np.sin(
      np.sqrt(np.abs(x - (y + 47.))))

  def zakharov(x):
    ii = np.arange(1, len(x) + 1, step=1, dtype=x.dtype)
    sum1 = (x**2).sum()
    sum2 = (0.5*ii*x).sum()
    answer = sum1+sum2**2+sum2**4
    return answer
  
  funs = dict(
    rosenbrock=rosenbrock,
    himmelblau=himmelblau,
    matyas=matyas,
    eggholder=eggholder,
    zakharov=zakharov
  )
  return funs[name]

class BfgsTest(test_util.JaxoptTestCase):

  def test_correctness(self):
    def fun(x, *args, **kwargs):  # Rosenbrock function.
      return sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1 - x[:-1])**2.0)

    x0 = jnp.zeros(2)
    bfgs = BFGS(fun=fun, tol=1e-3, stepsize=1e-3, maxiter=15)
    x1, _ = bfgs.run(x0)

    x2 = _bfgs(fun, x0, stepsize=1e-3, maxiter=15)

    self.assertArraysAllClose(x1, x2, atol=1e-5)

  @parameterized.product(linesearch=["backtracking", "zoom"])
  def test_binary_logreg(self, linesearch):
    X, y = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=2, n_informative=3, random_state=0
    )
    data = (X, y)
    fun = objective.binary_logreg

    w_init = jnp.zeros(X.shape[1])
    bfgs = BFGS(fun=fun, tol=1e-3, maxiter=500, linesearch=linesearch)
    # Test with keyword argument.
    w_fit, info = bfgs.run(w_init, data=data)

    # Check optimality conditions.
    self.assertLessEqual(info.error, 5e-2)

    # Compare against sklearn.
    w_skl = test_util.logreg_skl(X, y, 1e-6, fit_intercept=False,
                                 multiclass=False)
    self.assertArraysAllClose(w_fit, w_skl, atol=5e-2)

  @parameterized.product(
      linesearch=[
          "backtracking",
          "zoom",
          BacktrackingLineSearch(objective.multiclass_logreg_with_intercept),
      ],
      linesearch_init=["max", "current", "increase"]
  )
  def test_multiclass_logreg(self, linesearch, linesearch_init):
    data = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=3, n_informative=3, random_state=0
    )
    fun = objective.multiclass_logreg_with_intercept

    w_init = jnp.zeros((data[0].shape[1], 3))
    b_init = jnp.zeros(3)
    pytree_init = (w_init, b_init)

    bfgs = BFGS(
        fun=fun,
        tol=1e-3,
        maxiter=500,
        linesearch=linesearch,
        linesearch_init=linesearch_init,
    )
    # Test with positional argument.
    _, info = bfgs.run(pytree_init, data)

    # Check optimality conditions.
    self.assertLessEqual(info.error, 1e-2)

  @parameterized.product(n_iter=[10])
  def test_n_calls(self, n_iter):
    """Test whether the number of function calls
    is equal to the number of iterations + 1 in the
    no linesearch case, where the complexity is linear."""
    global N_CALLS
    N_CALLS = 0

    def fun(x, *args, **kwargs):  # Rosenbrock function.
      global N_CALLS
      N_CALLS += 1
      return sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1 - x[:-1])**2.0)

    x0 = jnp.zeros(2)
    bfgs = BFGS(fun=fun, tol=1e-7, maxiter=n_iter, jit=False, stepsize=1e-3)
    x1, _ = bfgs.run(x0)

    self.assertEqual(N_CALLS, n_iter + 1)


  @parameterized.product(
    fun_init_and_opt=[
      ('rosenbrock', onp.zeros(2), 0.),
      ('himmelblau', onp.ones(2), 0.),
      ('matyas', onp.ones(2) * 6., 0.),
      ('eggholder', onp.ones(2) * 100., None),  
    ],
  )
  def test_against_scipy(self, fun_init_and_opt):
    # Taken from previous jax tests
    # NOTE(vroulet): Unclear whether lbfgs or bfgs can find true minimum for
    # eggholder, but they seem to converge to the same solution with current 
    # implementation and initialization, which is a good check.

    # high precision for faithful checks
    tol = 1e-15 if jax.config.jax_enable_x64 else 1e-7
    # jaxopt_opts = dict(maxls=100) if jax.config.jax_enable_x64 else {}
    fun_name, x0, opt = fun_init_and_opt
    jnp_fun, onp_fun = get_fun(fun_name, jnp), get_fun(fun_name, onp)
    jaxopt_res = BFGS(jnp_fun, tol=tol).run(x0).params
    scipy_res = scipy_opt.minimize(onp_fun, x0, method='BFGS').x
    if fun_name == 'matyas':
    # scipy not good for matyas function, compare to true minimum, zero
      self.assertAllClose(jaxopt_res, jnp.zeros_like(jaxopt_res), check_dtypes=False)
    elif fun_name == 'eggholder' and jax.config.jax_enable_x64:
      # NOTE(vroulet): slight issue at high precision here
      self.assertAllClose(jnp_fun(jaxopt_res), onp_fun(scipy_res),
                          atol=1e-11, rtol=1e-11, check_dtypes=False)
    else:
      self.assertAllClose(jaxopt_res, scipy_res, check_dtypes=False)

    if opt is not None:
      self.assertAllClose(jnp_fun(jaxopt_res), opt, check_dtypes=False)

  @unittest.skipIf(not jax.config.jax_enable_x64, 'test requires X64')
  def test_zakharov(self):
    # Taken from previous jax tests
    # Function to steep to work without high precision (curiously lbfgs works)
    x0 = jnp.array([600.0, 700.0, 200.0, 100.0, 90.0, 1e4])
    fun = get_fun('zakharov', jnp)
    jaxopt_res = BFGS(fun, tol=1e-16, maxls=50).run(x0).params
    self.assertAllClose(jaxopt_res, jnp.zeros_like(jaxopt_res))
    self.assertAllClose(fun(jaxopt_res), 0.)

  def test_minimize_bad_initial_values(self):
    # Taken from previous jax tests
    # This test runs deliberately "bad" initial values to test that handling
    # of failed line search, etc. is the same across implementations
    initial_value = onp.array([92, 0.001])
    tol = 1e-6 if jax.config.jax_enable_x64 else 1e-3
    jnp_fun, onp_fun = get_fun('himmelblau', jnp), get_fun('himmelblau', onp)
    # Here reuqires higher number of linesearch for jaxopt
    jaxopt_res = BFGS(jnp_fun, tol=tol).run(initial_value).params
    scipy_res = scipy_opt.minimize(
        fun=onp_fun,
        jac=jax.grad(onp_fun),
        method='BFGS',
        x0=initial_value
    ).x
    # Scipy and jaxopt converge to different minima so check in function values
    self.assertAllClose(onp_fun(scipy_res), jnp_fun(jaxopt_res))
    self.assertAllClose(jnp_fun(jaxopt_res), jnp.asarray(0.))

  def test_steep(self):
    # Taken from previous jax tests
    # See jax related issue https://github.com/google/jax/issues/4594
    n = 2
    A = jnp.eye(n) * 1e4
    def fun(x):
      return jnp.mean((A @ x) ** 2)
    results = BFGS(fun).run(init_params=jnp.ones(n)).params
    self.assertAllClose(results, jnp.zeros(n))


if __name__ == '__main__':
  absltest.main()
